#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    This method computes metrics between a pair of graphs.
    To call this method, simply pass the reference graph, and the
    predicted graph (the one you want to make as much similar to the first one as
    possible), and all metrics will be computed.

    Use:
    >>> from random import random
    >>> target = nx.DiGraph()
    >>> target.add_nodes_from(['A', 'B', 'C', 'D', 'E'])
    >>> target.add_weighted_edges_from([\
        ('A', 'B', random()), ('B', 'D', random()),('C', 'B', random()),\
        ('D', 'E', random()), ('C', 'E', random())])

    >>> predicted = nx.DiGraph()
    >>> predicted.add_nodes_from(['A', 'B', 'C', 'D', 'E'])
    >>> predicted.add_weighted_edges_from([\
        ('A', 'B', random()), ('A', 'C', random()), ('E', 'A', random()),\
        ('E', 'B', random()), ('C', 'D', random())])

    >>> result = evaluate_graph(target, predicted)
"""

# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches

from dataclasses import dataclass
from typing import Dict, List, Optional, Union

import networkx as nx
import numpy as np
from networkx.linalg import adjacency_matrix
from sklearn.metrics import auc, precision_recall_curve

from causalexplain.metrics.SID import SID
from causalexplain.common import utils

# Code for metrics...
AnyGraph = Union[nx.Graph, nx.DiGraph]


@dataclass
class Metrics:
    """
        This class contains all the metrics computed by the evaluate_graph method.
    """

    Tp: int = 0
    Tn: int = 0
    Fn: int = 0
    Fp: int = 0
    precision: float = 0.0
    recall: float = 0.0
    aupr: float = 0.0
    f1: float = 0.0
    shd: float = 0.0
    sid: float = 0.0

    def __init__(
            self,
            Tp: int,
            Tn: int,
            Fn: int,
            Fp: int,
            precision: float,
            recall: float,
            aupr: float,
            f1: float,
            shd: int,
            sid: Dict[str, float]
    ):
        self.Tp = Tp
        self.Tn = Tn
        self.Fn = Fn
        self.Fp = Fp
        self.precision = precision
        self.recall = recall
        self.aupr = aupr
        self.f1 = f1
        self.shd = shd
        if self.Fn + self.Fp == 0:
            self.ishd = 0.0
        else:
            self.ishd = 1 - (self.shd / (self.Fn + self.Fp))
        self.sid = sid['sid']
        self.sid_lower = sid['sidLowerBound']
        self.sid_upper = sid['sidUpperBound']

    def to_dict(self) -> Dict[str, Union[int, float]]:
        """
        Convert the metrics to a dictionary.
        """
        dictionary = {}
        dictionary["Tp"] = self.Tp
        dictionary["Tn"] = self.Tn
        dictionary["Fn"] = self.Fn
        dictionary["Fp"] = self.Fp
        dictionary["precision"] = self.precision
        dictionary["recall"] = self.recall
        dictionary["f1"] = self.f1
        dictionary["aupr"] = self.aupr
        dictionary["shd"] = self.shd
        dictionary["sid"] = self.sid

        return dictionary

    def __str__(self):
        s = ""
        s += f"Predicted    1     0      Precision: {self.precision:.3g}\n"
        s += f"Actual     -----------    Recall...: {self.recall:.3g}\n"
        s += f"  1       |{self.Tp:^5d} {self.Fn:^5d}|   F1.......: {self.f1:.3g}\n"
        s += f"  0       |{self.Fp:^5d} {self.Tn:^5d}|   AuPR.....: {self.aupr:.3g}\n"
        s += f"           -----------    SHD......: {self.shd:.3g}\n"
        if self.sid_upper == self.sid_lower:
            s += f"                          SID......: {self.sid:.3g}\n"
        else:
            s += f"                          SID......: {self.sid:.3g} [{self.sid_lower}..{self.sid_upper}]\n"
        return s

    def matplotlib_repr(self):
        """
        Generates a formatted string representation of the metrics for display
        in a matplotlib plot.

        Returns:
            str: The formatted string representation of the metrics.
        """
        s_Tp = f"{self.Tp:^5d}"
        s_Tn = f"{self.Tn:^5d}"
        s_Fp = f"{self.Fp:^5d}"
        s_Fn = f"{self.Fn:^5d}"
        s_Tp = s_Tp.replace(" ", r"\ ")
        s_Tn = s_Tn.replace(" ", r"\ ")
        s_Fp = s_Fp.replace(" ", r"\ ")
        s_Fn = s_Fn.replace(" ", r"\ ")

        s = r"> Predicted\ \ \ 1\ \ \ \ \ 0" + "\n"
        s += r"> Actual\ \ \ \ -----------" + "\n"
        s += r">\ \ 1\ \ \ \ \ \ \ |" + s_Tp + r"\ " + s_Fn + "|" + "\n"
        s += r">\ \ 0\ \ \ \ \ \ \ |" + s_Fp + r"\ " + s_Tn + "|" + "\n"
        s += r">\ \ \ \ \ \ \ \ \ \ \ -----------" + "\n"
        s += f"> Precision: {self.precision:.3g}" + "\n"
        s += f"> Recall...: {self.recall:.3g}" + "\n"
        s += f"> F1.......: {self.f1:.3g}" + "\n"
        s += f"> AuPR.....: {self.aupr:.3g}" + "\n"
        s += f"> SHD......: {self.shd:.1f}" + "\n"
        s += f"> SID......: {self.sid}"

        return s


def _shallow_copy(G: AnyGraph):
    """
    This method returns a shallow copy of the graph G.

    Parameters
    ----------
    G : AnyGraph
        The graph to copy.

    Returns
    -------
    AnyGraph
        The shallow copy of the graph.
    """
    G_new = nx.DiGraph()
    G_new.add_edges_from(G.edges())
    return G_new


def evaluate_graph(
    ground_truth: AnyGraph,
    predicted_graph: AnyGraph,
    feature_names: Optional[List[str]] = None,
    threshold: float = 0.0,
    absolute: bool = False,
    double_for_anticausal: bool = True,
) -> Optional[Metrics]:
    """
        This method computes metrics between a pair of graphs: the ground truth and
        the predicted graph. To call this method, simply pass the reference graph, and
        the predicted graph (the one you want to make as much similar to the first one
        as possible), and all metrics will be computed.

        Parameters
        ----------
        ground_truth : AnyGraph
            The ground truth graph.
        predicted_graph : AnyGraph
            The predicted graph.
        feature_names : Optional[List[str]], optional
            The list of feature names, by default None
        threshold : float, optional
            The threshold to use for the precision-recall curve, by default 0.0
        absolute : bool, optional
            Whether to use the absolute value of the weights, by default False
        double_for_anticausal : bool, optional
            Whether to double the weights of anticausal edges, by default True

        Returns
        -------
        Metrics
            The metrics object containing all the metrics: Tp, Tn, Fn, Fp, precision,
            recall, AuPR, F1, SHD, ISHD, UMI, OMI.
    """
    if ground_truth is None:
        return None

    # If one graphs is directed and the other is not, then consider them both as
    # undirected.
    if (nx.is_directed(ground_truth) and not nx.is_directed(predicted_graph)) or (
            not nx.is_directed(ground_truth) and nx.is_directed(
                predicted_graph)
    ):
        truth = ground_truth.to_undirected() if nx.is_directed(
            ground_truth) else _shallow_copy(ground_truth)
        predicted = predicted_graph.to_undirected() if nx.is_directed(
            predicted_graph) else _shallow_copy(ground_truth)
    else:
        truth = ground_truth.copy()
        predicted = predicted_graph.copy()

    # Checking whether both contain the same nr of nodes.
    # If not, add isolated nodes to the 'g' graph
    missing_nodes = list(set(ground_truth.nodes()) -
                         set(predicted_graph.nodes()))
    if len(missing_nodes) != 0:
        predicted.add_nodes_from(missing_nodes)

    if feature_names is None:
        feature_names = list(predicted_graph.nodes())

    metrics = dict()
    metrics["_Gm"] = _adjacency(truth, feature_names, threshold, absolute)
    metrics["_gm"] = _adjacency(predicted, feature_names, threshold, absolute)
    if _is_weighted(predicted):
        metrics["_preds"] = _weighted_adjacency(
            predicted, feature_names, threshold, absolute
        )
    else:
        metrics["_preds"] = metrics["_gm"].copy()
    metrics["_double_for_anticausal"] = double_for_anticausal

    metrics["Tp"], metrics["Tn"], metrics["Fp"], metrics["Fn"] = _conf_mat(
        truth, predicted, feature_names
    )

    metrics["precision"] = _precision(metrics)
    metrics["recall"] = _recall(metrics)
    metrics["f1"] = _f1(metrics)
    metrics["aupr"] = _aupr(metrics)
    metrics["SHD"] = _SHD(metrics)
    # Cross check if the number of nodes in the predicted graph is the same as the
    # number of nodes in the ground truth graph
    if predicted_graph.number_of_nodes() < ground_truth.number_of_nodes():
        # Determine what are the nodes missing in the predicted graph
        missing_nodes = list(set(ground_truth.nodes()) -
                             set(predicted_graph.nodes()))
        # Create a copy of the predicted graph
        predicted_graph_copy = predicted_graph.copy()
        # Add the missing nodes to the predicted graph
        predicted_graph_copy.add_nodes_from(missing_nodes)
    #     predicted_graph_array = nx.to_numpy_array(predicted_graph_copy)
    # else:
    #     predicted_graph_array = nx.to_numpy_array(predicted_graph)

    true_adj = utils.graph_to_adjacency(ground_truth, feature_names)
    true_adj = true_adj.astype(np.int8)
    est_adj = utils.graph_to_adjacency(predicted_graph, feature_names)
    est_adj = est_adj.astype(np.int8)
    metrics["SID"] = SID(trueGraph=true_adj, estGraph=est_adj)

    return Metrics(metrics["Tp"], metrics["Tn"], metrics["Fn"], metrics["Fp"],
                   metrics["precision"], metrics["recall"], metrics["aupr"],
                   metrics["f1"], metrics["SHD"], metrics["SID"])


def _is_weighted(G: AnyGraph) -> bool:
    """
    Returns if a graph is weighted by checking if list of edges has data

    Arguments:
        - G (Graph or DiGraph): a graph or digraph

    Returns:
        True if edges have no data.
    """
    ddicts = map(lambda t: bool(t[2]), list(G.edges().data()))
    return any(ddicts)


def _binary_adj_matrix(
        G: AnyGraph,
        order: Optional[List] = None,
        threshold: float = 0.0,
        absolute: bool = False
) -> np.ndarray:
    """
    Returns a binary adjacency matrix from a weighted adjacency matrix. If the
    values in the adjacency matrix are greater than the threshold (default 0.0)
    then that value is transformed into 1.0.

    Arguments:
        - G (Graph or DiGraph): Graph or Digraph
        - threshold (float): Min value of weight to be valued as 1 in the binary
            matrix.
        - absolute (bool): Whether performing the comparison of weights against
            the threshold using absolute value. Default is false.

    Returns:
        np.ndarray with the binary version of the weights.
    """
    if order is None:
        order = sorted(list(G.nodes()))
    m = np.zeros((len(order), len(order)))
    for u, v in G.edges():
        w = G.get_edge_data(u, v)['weight']
        if w is not None:
            if absolute:
                w = np.abs(w)
            if w >= threshold:
                m[order.index(u), order.index(v)] = 1

    return m.astype(np.int16)


def _adjacency(
        G: AnyGraph,
        order: Optional[List] = None,
        threshold=0.0,
        absolute=False) -> np.ndarray:
    """
    Retrieves the adjacency matrix from the graph using NetworkX method if the
    graph is not weighted. This method produces a matrix with 1/0 values only
    since it is used to compute metrics.

    Arguments:
        - G (Graph or DiGraph): the graph to extract the adj. matrix from.
        - order (list): The order of nodes to be used in the adjacency matrix. If
            none specified the node list will be sorted in ascending order.
        - threshold (float): The minimum weight for an edge to be considered as
            present in the adjacency matrix. Default is 0.0.
        - absolute (bool): Default is false. Whether to consider absoute values
            when comparing edge weights agains the threshold.

    Returns:
        numpy.matrix: An adjacency matrix formed by 0s and 1s where the order of
            the nodes is given as argument to the class, and the minimum weight
            must be above the threshold also specified.
    """
    if order is None:
        order = sorted(list(G.nodes()))
    if nx.is_weighted(G) and all(
            [x[-1]['weight'] is not None for x in G.edges(data=True)]):
        result = _binary_adj_matrix(G, order, threshold, absolute)
    else:
        # Scipy adjacency sometimes throws an exception based on type.
        # result = adjacency_matrix(G, nodelist=order).todense()
        nodeset = set(order)
        if nodeset - set(G):
            # delete from order any nodes not in G
            order = [n for n in order if n in G.nodes()]
        adj_matrix = nx.to_numpy_array(G, nodelist=order)
        adj_matrix[np.isnan(adj_matrix)] = 0
        result = adj_matrix

    return result


def _weighted_adjacency(
        G: AnyGraph,
        order: Optional[List] = None,
        threshold=0.0,
        absolute=False) -> np.ndarray:
    """
    Retrieves the adjacency matrix from the graph using NetworkX method if the
    graph is not weighted. This method produces a matrix with 1/0 values only

    Arguments:
        - G (Graph or DiGraph): the graph to extract the adj. matrix from.
        - order (list): The order of nodes to be used in the adjacency matrix. If
            None, the order of the nodes in the graph will be used.
        - threshold (float): The minimum weight to be considered in the matrix.
            Default is 0.0.
        - absolute (bool): Whether performing the comparison of weights against
            the threshold using absolute value. Default is false.

    Returns:
        numpy.matrix: An adjacency matrix formed by 0s and 1s where the order of
            the nodes is given as argument to the class, and the minimum weight
            must be above the threshold also specified.
    """
    if order is None:
        order = sorted(list(G.nodes()))
    F = G.copy()

    def value(x): return abs(x) if absolute else x

    adj_matrix = np.zeros((len(order), len(order)))
    for u, v in F.edges():
        # check if F.get_edge_data(u, v) is not None and contains a key 'weight'
        if 'weight' not in F.get_edge_data(u, v):
            w = 1
        else:
            w = value(F.get_edge_data(u, v)['weight'])
        if w is not None:
            if w >= threshold:
                adj_matrix[order.index(u), order.index(v)] = w
                if not G.is_directed():
                    adj_matrix[order.index(v), order.index(u)] = w
        else:
            adj_matrix[order.index(u), order.index(v)] = 0
            if not G.is_directed():
                adj_matrix[order.index(v), order.index(u)] = 0


    return adj_matrix


def _intersect_matrices(mat1: np.ndarray, mat2: np.ndarray) -> np.ndarray:
    """
    Computes the intersection of two matrices.

    Parameters:
        mat1: (np.matrix) the first matrix
        mat2: (np.matrix) the second matrix

    Returns:
        np.ndarray: A new matrix with the values in common between the two matrices.
    """
    if not (mat1.shape == mat2.shape):
        raise ValueError("Both graphs must have the same number of nodes")
    mat_intersect = np.where((mat1 == mat2), mat1, 0)

    return mat_intersect


def _positive(matrix: np.ndarray) -> np.ndarray:
    """
    Returns a matrix where negative values are converted to zero,
    and positive values remain the same
    """
    return (matrix > 0).astype(int) * matrix


def _negative(matrix: np.ndarray) -> np.ndarray:
    """
    Returns a matrix where positive values are converted to zero,
    and negative values remain the same.
    """
    return (matrix < 0).astype(int) * matrix



def _conf_mat(truth, est, feature_names):
    if truth.is_directed() and est.is_directed():
        return _conf_mat_directed(truth, est, feature_names)
    else:
        return _conf_mat_undirected(truth, est, feature_names)


def _conf_mat_directed(truth, est, feature_names):
    """
    Computes the confusion matrix for two directed graphs. This method is
    currently only used for directed graphs.

    Arguments:
        truth: (nx.DiGraph) the ground truth graph
        est: (nx.DiGraph) the estimated graph

    Returns:
        Tp: (int) number of true positives
        Tn: (int) number of true negatives
        Fp: (int) number of false positives
        Fn: (int) number of false negatives
    """
    assert truth.is_directed(), "true graph must be directed"
    assert est.is_directed(), "estimated graph must be directed"

    def is_arrow(G, u, v):
        if u not in G.nodes() or v not in G.nodes():
            return False
        return G.has_edge(u, v)  # and not G.has_edge(v, u)

    num_elems = len(feature_names)

    truePositives = np.zeros(
        (num_elems, num_elems)).astype(int)
    estPositives = np.zeros(
        (num_elems, num_elems)).astype(int)
    for i, u in enumerate(feature_names):
        for j, v in enumerate(feature_names):
            if is_arrow(truth, u, v):
                truePositives[i, j] = 1
            if is_arrow(est, u, v):
                estPositives[i, j] = 1

    zeros = np.zeros((num_elems, num_elems))

    Tp = (np.minimum(truePositives == estPositives, truePositives)).sum()
    Tn = (truePositives == estPositives).sum() - Tp - num_elems
    Tn = 0 if Tn < 0 else Tn

    Fp = int((np.maximum(estPositives - truePositives, zeros)).sum())
    Fn = int((np.maximum(truePositives - estPositives, zeros)).sum())

    return Tp, Tn, Fp, Fn


def _conf_mat_undirected(truth, est, feature_names):
    """
    Computes the confusion matrix for two undirected graphs. This method is
    currently only used for undirected graphs.

    Arguments:
        truth: (nx.Graph) the ground truth graph
        est: (nx.Graph) the estimated graph

    Returns:
        Tp: (int) number of true positives
        Tn: (int) number of true negatives
        Fp: (int) number of false positives
        Fn: (int) number of false negatives
    """
    assert not truth.is_directed(), "true graph must be undirected"
    assert not est.is_directed(), "estimated graph must be undirected"

    def is_edge(G, u, v):
        if u not in G.nodes() or v not in G.nodes():
            return False
        return G.has_edge(u, v)

    truePositives = np.zeros(
        (len(feature_names), len(feature_names))).astype(int)
    estPositives = np.zeros(
        (len(feature_names), len(feature_names))).astype(int)
    for i, u in enumerate(feature_names):
        for j, v in enumerate(feature_names):
            if is_edge(truth, u, v):
                truePositives[i, j] = 1
                truePositives[j, i] = 1
            if is_edge(est, u, v):
                estPositives[i, j] = 1
                estPositives[j, i] = 1

    zeros = np.zeros((len(feature_names), len(feature_names)))

    # Excluir las diagonales del conteo
    eq = (truePositives == estPositives)
    np.fill_diagonal(eq, False)

    Tp = int((np.minimum(eq, truePositives).sum() / 2))
    Tn = int(eq.sum() / 2) - Tp
    Fp = int((np.maximum(estPositives - truePositives, zeros)).sum() / 2)
    Fn = int((np.maximum(truePositives - estPositives, zeros)).sum() / 2)

    return Tp, Tn, Fp, Fn


def _confusion_matrix(G, metrics):
    def comp(M):
        return 1 - M

    # These are the intersection between the adjacency matrices. The second one
    # is between the complementary matrices
    G_n_g = _intersect_matrices(metrics["_Gm"], metrics["_gm"])
    cG_n_cg = _intersect_matrices(comp(metrics["_Gm"]), comp(metrics["_gm"]))
    N = metrics["_Gm"].shape[0]
    if isinstance(G, nx.DiGraph):
        _Tp = int(G_n_g.sum())
        _Tn = (
            cG_n_cg[np.triu_indices(n=N, k=+1)].sum()
            + cG_n_cg[np.tril_indices(n=N, k=-1)].sum()
        )
        _Fp = int(_negative(metrics["_Gm"] - metrics["_gm"]).sum())
        _Fn = int(_negative(metrics["_gm"] - metrics["_Gm"]).sum())
    elif isinstance(G, nx.Graph):
        _Tp = int(G_n_g.sum() / 2)
        _Tn = cG_n_cg[np.triu_indices(n=N, k=1)].sum()
        _Fp = int(_negative(metrics["_Gm"] - metrics["_gm"]).sum() / 2)
        _Fn = int(_negative(metrics["_gm"] - metrics["_Gm"]).sum() / 2)
    else:
        raise TypeError("Only networkx Graph and DiGraph graphs supported")
    return _Tp, _Tn, _Fn, _Fp


def _precision(metrics):
    """
    Precision = True Positives / (True Positives + False Positives)

    Arguments:
        metrics (dict): A dictionary containing the following keys:
            - Tp (int): The number of true positives.
            - Fp (int): The number of false positives.

    Returns:
        float: The precision score.
    """
    if (metrics["Tp"] + metrics["Fp"]) == 0:
        return 0
    else:
        return metrics["Tp"] / (metrics["Tp"] + metrics["Fp"])


def _recall(metrics):
    """
    Recall = True Positives / (True Positives + False Negatives)

    Arguments:
        metrics (dict): A dictionary containing the following keys:
            - Tp (int): The number of true positives.
            - Fn (int): The number of false negatives.

    Returns:
        float: The recall score.
    """
    if (metrics["Tp"] + metrics["Fn"]) == 0.:
        return 0
    else:
        return metrics["Tp"] / (metrics["Tp"] + metrics["Fn"])


def _aupr(metrics):
    """
    Area Under the Precision-Recall Curve (AUPR)

    Arguments:
        metrics (dict): A dictionary containing the following keys:
            - _Gm (numpy.ndarray): The ground truth adjacency matrix.
            - _preds (numpy.ndarray): The predicted adjacency matrix.

    Returns:
        float: The AUPR score.
    """
    true_labels = np.squeeze(np.asarray(metrics["_Gm"].ravel()))
    predicted = np.squeeze(np.asarray(metrics["_preds"].ravel()))
    try:
        _sk_precision, _sk_recall, _ = precision_recall_curve(
            true_labels, predicted)
    except ValueError:
        return 0.0
    return auc(_sk_recall, _sk_precision)


def _f1(metrics):
    """
    F1 = 2 * (Precision * Recall) / (Precision + Recall)

    Arguments:
        metrics (dict): A dictionary containing the following keys:
            - precision (float): The precision score.
            - recall (float): The recall score.

    Returns:
        float: The F1 score.
    """
    if metrics["recall"] + metrics["precision"] == 0.0:
        return 0.0
    else:
        return (
            2
            * (metrics["recall"] * metrics["precision"])
            / (metrics["recall"] + metrics["precision"])
        )


def _SHD(metrics):
    r"""
    Compute the Structural Hamming Distance.
    The Structural Hamming Distance (SHD) is a standard distance to compare
    graphs by their adjacency matrix. It consists in computing the difference
    between the two (binary) adjacency matrixes: every edge that is either
    missing or not in the target graph is counted as a mistake. Note that
    for directed graph, two mistakes can be counted as the edge in the wrong
    direction is false and the edge in the good direction is missing ; the
    `double_for_anticausal` argument accounts for this remark. Setting it to
    `False` will count this as a single mistake.

    Returns:
        int: Structural Hamming Distance (int).
            The value tends to zero as the graphs tend to be identical.
    """
    diff = np.abs(metrics["_Gm"] - metrics["_gm"])
    if metrics["_double_for_anticausal"]:
        _shd = np.sum(diff)
    else:
        diff = diff + diff.transpose()
        diff[diff > 1] = 1
        _shd = np.sum(diff) / 2

    return int(_shd)



import numpy as np

from causalexplain.explainability.regression_quality import RegQuality


def test_predict_intersects_gamma_and_mad():
    scores = [1.0, 1.0, 10.0]
    result = RegQuality.predict(scores, gamma_shape=1, gamma_scale=1, threshold=0.05)
    assert result == {2}


def test_gamma_and_mad_helpers_handle_small_sets():
    scores = np.array([0.1, 0.2, 0.3])
    gamma_idx = RegQuality._gamma_criteria(scores, gamma_shape=1, gamma_scale=1, threshold=10)
    assert gamma_idx == {0, 1, 2}

    mad_idx = RegQuality._mad_criteria(scores, verbose=True)
    assert mad_idx == set()



"""
FCI algorithm.

A graph learner which wraps the implementation of the FCI algorithm
in the (c) causallearn library (https://github.com/py-why/causal-learn).

"""

# pylint: disable=E1101:no-member
# pylint: disable=W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=W0106:expression-not-assigned
# pylint: disable=C0103:invalid-name, C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments, R0902:too-many-instance-attributes
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=R1702:too-many-branches

import networkx as nx
import numpy as np
import pandas as pd
from causallearn.search.ConstraintBased.FCI import fci

from causalexplain.common import utils
from causalexplain.metrics.compare_graphs import evaluate_graph


class FCI:
    """
    A graph learner which wraps the implementation of the FCI algorithm
    in (c) causallearn library (https://github.com/py-why/causal-learn).
    """

    def __init__(
            self,
            name: str,
            independence_test_method: str = "fisherz",
            alpha: float = 0.05,
            depth: int = -1,
            max_path_length: int = -1,
            verbose: bool = False,
            background_knowledge=None,
            show_progress: bool = True,
            node_names=None,
            causes_only=True):
        """
        Initialize the FCI algorithm creating an FCI learner.

        Parameters
        ----------
        name: str, name of the experiment
        independence_test_method: str, name of the function of the independence 
            test method being used: [fisherz, chisq, gsq, kci]
            - fisherz: Fisher's Z conditional independence test
            - chisq: Chi-squared conditional independence test
            - gsq: G-squared conditional independence test
            - kci: Kernel-based conditional independence test
        alpha: float, desired significance level of independence tests 
            (p_value) in (0,1)
        depth: The depth for the fast adjacency search, or -1 if unlimited
        max_path_length: the maximum length of any discriminating path, 
            or -1 if unlimited.
        verbose: True is verbose output should be printed or logged
        background_knowledge: background knowledge
        causes_only: bool, if True, only causes are returned
            (default: True), by filtering the CPDAG and considering only
            the edges indicating a causal relationship.
        """
        super().__init__()
        self.name = name
        self.independence_test_method = independence_test_method
        self.alpha = alpha
        self.depth = depth
        self.max_path_length = max_path_length
        self.verbose = verbose
        self.background_knowledge = background_knowledge
        self.prog_bar = show_progress
        self.node_names = node_names
        self.causes_only = causes_only

    def fit_predict(self,
                    X: pd.DataFrame,
                    X_test: pd.DataFrame = None,
                    ref_graph: nx.DiGraph = None):
        """
        Fits the model to the data and returns predictions.

        Parameters
        ----------
        X (pd.DataFrame): The input data to fit the model on.
        X_test (pd.DataFrame): The input data to fit the model on.
        ref_graph (nx.DiGraph): The reference graph, or ground truth.
        causes_only: bool, if True, only causes are returned
            (default: True), by filtering the CPDAG and considering only
            the edges indicating a causal relationship.

        Returns
        -------
        graph : a GeneralGraph object, where graph.graph[j,i]=1 and 
            graph.graph[i,j]=-1 indicates  i --> j ,
            graph.graph[i,j] = graph.graph[j,i] = -1 indicates i --- j,
            graph.graph[i,j] = graph.graph[j,i] = 1 indicates i <-> j,
            graph.graph[j,i]=1 and graph.graph[i,j]=2 indicates  i o-> j.
        edges : list
            Contains graph's edges properties.
            If edge.properties have the Property 'nl', then there is no latent 
                confounder. Otherwise, there are possibly latent confounders.
            If edge.properties have the Property 'dd', then it is definitely 
                direct. Otherwise, it is possibly direct.
            If edge.properties have the Property 'pl', then there are 
                possibly latent confounders. Otherwise, there is no 
                latent confounder.
            If edge.properties have the Property 'pd', then it is 
            possibly direct. Otherwise, it is definitely direct.
        """
        graph, edges = fci(
            X.values,
            indep_test_method=self.independence_test_method,
            alpha=self.alpha,
            depth=self.depth,
            max_path_length=self.max_path_length,
            verbose=False,
            show_progress=False,
            background_knowledge=self.background_knowledge,
            node_names=self.node_names
        )

        if self.causes_only:
            adjacency_matrix = self.filter_causes_only(
                adj_matrix=graph.graph)
            self.dag = utils.graph_from_adjacency(
                adjacency_matrix, node_labels=list(X.columns))
        else:
            self.dag = utils.graph_from_adjacency(
                graph.graph, node_labels=list(X.columns))

        self.metrics = evaluate_graph(ref_graph, self.dag)

        return self

    def filter_causes_only(self, adj_matrix):
        """
        Filter the adjacency matrix to include only causal relationships.

        This method processes the input adjacency matrix and creates a new 
        matrix that only includes direct causal relationships 
        (i.e., where i causes j).

        Parameters:
        -----------
        adj_matrix : numpy.ndarray
            The input adjacency matrix to be filtered.

        Returns:
        --------
        numpy.ndarray
            A new adjacency matrix with only causal relationships preserved.
        """
        result_matrix = np.zeros_like(adj_matrix)

        # Iterate over each pair (i, j) in the matrix
        for i in range(adj_matrix.shape[0]):
            for j in range(adj_matrix.shape[1]):
                # If G[i, j] = -1 and G[j, i] = 1, set result_matrix[i, j] to 1
                if (adj_matrix[i, j] == -1 and adj_matrix[j, i] == 1) \
                        or (adj_matrix[i, j] == 2 and adj_matrix[j, i] == 1):
                    result_matrix[i, j] = 1

        return result_matrix


def main(dataset_name,
         input_path="/Users/renero/phd/data/sachs/",
         output_path="/Users/renero/phd/output/RC4/sachs/compared/",
         save=False,
         **kwargs):
    """
    Create a call to FCI with a sample dataset.
    """
    data = pd.read_csv(f"{input_path}{dataset_name}.csv")
    ref_graph = utils.graph_from_dot_file(f"{input_path}{dataset_name}.dot")

    fci = FCI(name=dataset_name, **kwargs)
    fci.fit_predict(X=data, ref_graph=ref_graph)

    if fci.dag:
        for edge in fci.dag.edges():
            print(edge)
    else:
        for edge in fci.pag.edges():
            print(edge)
    if fci.metrics:
        print(fci.metrics)
    else:
        print("No metrics available")

    # if save:
    #     where_to = utils.save_experiment(rex.name, output_path, rex)
    #     print(f"Saved '{rex.name}' to '{where_to}'")


# Create a call to FCI with a sample dataset.
if __name__ == "__main__":
    main("sachs", alpha=0.75, depth=9)
    # main("rex_generated_linear_1", njobs=1)



import logging
import math

import numpy as np
import pytorch_lightning as pl
import torch
from torch import Tensor, nn
from torch.autograd import Variable
from torch.distributions import Categorical
from torch.nn.functional import softmax
from torch.optim import Adam
from torch.optim.lr_scheduler import ReduceLROnPlateau

from ._loss import MMDLoss
from ..common import utils

logging.getLogger('pytorch_lightning').setLevel(logging.ERROR)
torch_log = logging.getLogger("pytorch_lightning")
torch_log.propagate = False
torch_log.setLevel(logging.ERROR)


ONEOVERSQRT2PI = 1.0 / math.sqrt(2 * math.pi)
DEVICE = "cpu"


class MLP(pl.LightningModule):

    device = utils.select_device(DEVICE)

    class Block(nn.Module):
        """The main building block of `MLP`."""

        def __init__(
                self,
                d_in: int,
                d_out: int,
                activation,
                bias: bool,
                dropout: float,
                device):
            super().__init__()
            self.linear = nn.Linear(d_in, d_out, bias)
            self.activation = activation
            self.dropout = nn.Dropout(dropout)
            self.device = device  # Set device as an attribute

        def forward(self, x: Tensor) -> Tensor:
            return self.dropout(self.activation(self.linear(x)))

    def __init__(
            self,
            input_size,
            layers_dimensions,
            activation,
            batch_size,
            lr,
            loss,
            dropout):
        super(MLP, self).__init__()
        # Log args to Tensorboard.
        self.save_hyperparameters()

        # Net params
        self.input_size = input_size
        self.batch_size = batch_size
        self.learning_rate = lr
        self.dropout = dropout
        self.hidden_layers = len(layers_dimensions)

        if activation == "relu":
            self.activation = nn.ReLU()
        elif activation == "gelu":
            self.activation = nn.GELU()
        elif activation == "selu":
            self.activation = nn.SELU()
        elif activation == "tanh":
            self.activation = nn.Tanh()
        elif activation == "linear":
            self.activation = nn.Identity()
        elif activation == "sigmoid":
            self.activation = nn.Sigmoid()
        else:
            raise ValueError(
                "Unknown activation function (only 'relu' or 'selu').")

        if loss == "mse":
            self.loss_fn = nn.MSELoss()
        elif loss == "mae":
            self.loss_fn = nn.L1Loss()
        elif loss == "mmd":
            self.loss_fn = MMDLoss()  # mmd_type="quadratic")
        elif loss == "binary_crossentropy":
            self.loss_fn = nn.BCEWithLogitsLoss()
        elif loss == "crossentropy":
            self.loss_fn = nn.CrossEntropyLoss()
        else:
            raise ValueError("Unknown loss function.")

        if isinstance(self.dropout, float):
            dropouts = [self.dropout] * len(layers_dimensions)
            assert len(layers_dimensions) == len(dropouts)
        self.net = nn.Sequential(
            *[
                MLP.Block(
                    d_in=layers_dimensions[i - 1] if i else self.input_size,
                    d_out=d,
                    activation=self.activation,
                    bias=True,
                    dropout=dropout,
                    device=self.device
                )
                for i, (d, dropout) in enumerate(zip(layers_dimensions, dropouts))
            ]
        )
        self.head = nn.Linear(
            layers_dimensions[-1] if layers_dimensions else input_size, 1)

        if isinstance(self.activation, nn.SELU):
            for param in self.net.parameters():
                # biases zero
                if len(param.shape) == 1:
                    nn.init.constant_(param, 0)
                else:
                    nn.init.kaiming_normal_(
                        param, mode='fan_in', nonlinearity='linear')

    def forward(self, x):
        noise = torch.randn(x.shape[0], 1, device="cpu")
        # noise = torch.randn(x.shape[0], 1, device=x.device)
        X = torch.cat([x, noise], dim=1)
        y = self.net(X)
        y = self.head(y)
        return y

    def predict_step(self, batch, batch_idx, **kwargs):
        x, y = batch
        y_hat = self(x)
        return y_hat

    def training_step(self, batch, batch_idx):
        x, y = batch
        yhat = self(x)
        loss = self.loss_fn(yhat, y)
        self.log("train_loss", loss, on_step=False, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        yhat = self(x)
        loss = self.loss_fn(yhat, y)
        self.log("val_loss", loss, on_step=False, on_epoch=True)
        return loss

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer

    # To allow compatibility with SHAP Explainers different from DeepExplainer and
    # GradientExplainer. Taken from:
    # https://jamesmccaffrey.wordpress.com/2022/10/11/an-example-of-using-the-shap-library-for-a-pytorch-model/
    def predict(self, x):
        # x is numpy not tensor, return is numpy
        xx = torch.tensor(x, dtype=torch.float32, device=self.device)
        with torch.no_grad():
            probs = torch.exp(self.forward(xx))
        return probs.numpy()


class DFF(pl.LightningModule):
    def __init__(self, input_size, hidden_size, batch_size, lr, loss):
        super(DFF, self).__init__()
        # Log args to Tensorboard.
        self.save_hyperparameters()

        # Net params
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.batch_size = batch_size
        self.learning_rate = lr

        # Create the net
        self.approximate = nn.Sequential(
            nn.Linear(self.input_size, self.hidden_size),
            nn.Sigmoid(),
            nn.Linear(self.hidden_size, 1),
        )

        # Loss
        if loss == "mse":
            self.loss_fn = nn.MSELoss()
        elif loss == "mae":
            self.loss_fn = nn.L1Loss()
        elif loss == "mmd":
            self.loss_fn = MMDLoss(mmd_type="quadratic")
        else:
            raise ValueError("Unknown loss function.")

    def forward(self, x):
        # this_device = 'mps' if self.on_gpu else 'cpu'
        noise = torch.randn(x.shape[0], 1, device=self.device)
        X = torch.cat([x, noise], dim=1)
        y = self.approximate(X)
        return y

    def predict_step(self, batch, batch_idx, **kwargs):
        x, y = batch
        y_hat = self(x)
        return y_hat

    def training_step(self, batch, batch_idx):
        x, y = batch
        yhat = self(x)
        loss = self.loss_fn(yhat, y)
        self.log("train_loss", loss, on_step=False, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        yhat = self(x)
        loss = self.loss_fn(yhat, y)
        self.log("val_loss", loss, on_step=False, on_epoch=True)
        # return loss

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        # optimizer = torch.optim.SGD(self.parameters(), lr=learning_rate)
        return optimizer


class MDN(pl.LightningModule):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_gaussians: int,
        lr: float,
        batch_size: int,
        loss_function: str = "loglikelihood",
    ):
        """
        Init function for the MDN

        Arguments:
            input_size (int): the number of dimensions in the input
            hidden_size (int): the number of dimensions in the hidden layer
            num_gaussians (int): the number of Gaussians per output dimensions
            lr (float): learning rate
            batch_size (int): Batch size.
            loss_function (str): Loss function can be either 'loglikelihood' or 'mmd'
                for Maximal Mean Discrepancy
        Input:
            minibatch (BxD): B is the batch size and D is the number of input
                dimensions.
        Output:
            (pi, sigma, mu) (BxG, BxGxO, BxGxO): B is the batch size, G is the
                number of Gaussians, and O is the number of dimensions for each
                Gaussian. Pi is a multinomial distribution of the Gaussians. Sigma
                is the standard deviation of each Gaussian. Mu is the mean of each
                Gaussian.
        """

        super(MDN, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_gaussians = num_gaussians
        self.lr = lr
        self.batch_size = batch_size
        self.loss_fn = loss_function

        self.save_hyperparameters()

        self.mdn = nn.Sequential(nn.Linear(input_size, hidden_size), nn.Tanh())
        self.pi = nn.Linear(hidden_size, num_gaussians)
        self.sigma = nn.Sequential(
            nn.Linear(hidden_size, num_gaussians), nn.ELU())
        self.mu = nn.Linear(hidden_size, num_gaussians)

    def configure_optimizers(self):
        optimizer = Adam(self.parameters(), lr=(self.lr or self.learning_rate))
        lr_scheduler = ReduceLROnPlateau(
            optimizer,
        )
        return {
            "optimizer": optimizer,
            "lr_scheduler": lr_scheduler,
            "monitor": "val_loss",
        }

    def training_step(self, batch, batch_idx):
        loss = self.common_step(batch)
        self.log("train_loss", loss, on_step=False, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        if batch[0].size()[0] == self.batch_size:
            loss = self.common_step(batch)
        else:
            loss = 0.0
        self.log("val_loss", loss, on_step=False, on_epoch=True)

    def common_step(self, batch):
        x, y = batch
        # x = self.add_noise(x)
        pi, sigma, mu = self(x)
        if self.loss_fn == "loglikelihood":
            loss = self.mdn_loss(pi, sigma, mu, y)
        else:
            y_hat = Variable(self.g_sample(pi, sigma, mu), requires_grad=True)
            loss = self.mmd_loss(y.float(), y_hat.float(), "multiscale")
        return loss

    def forward(self, x):
        # this_device = 'mps' if self.on_gpu else 'cpu'
        noise = torch.randn(x.shape[0], 1, device=self.device)
        X = torch.cat([x.to_device(self.device), noise], dim=1)
        z_h = self.mdn(X)
        pi = softmax(self.pi(z_h), -1)
        sigma = torch.exp(self.sigma(z_h))
        mu = self.mu(z_h)
        return pi, sigma, mu

    def mdn_loss(self, pi, sigma, mu, y):
        """
        Calculates the error, given the MoG parameters and the target
        The loss is the negative log likelihood of the data given the MoG
        parameters.
        """
        result = self.gaussian_probability(y, mu, sigma) * pi
        result = torch.sum(result, dim=1)
        result = -torch.log(result)
        return torch.mean(result)

    @staticmethod
    def gaussian_probability(y, mu, sigma):
        # make |mu|=K copies of y, subtract mu, divide by sigma
        result = (y.expand_as(mu) - mu) * torch.reciprocal(sigma)
        result = -0.5 * (result * result)
        return (torch.exp(result) * torch.reciprocal(sigma)) * ONEOVERSQRT2PI

    @staticmethod
    def mmd_loss(x, y, kernel):
        """
        https://www.kaggle.com/onurtunali/maximum-mean-discrepancy

        Emprical maximum mean discrepancy. The lower the result,
        the more evidence that distributions are the same.

        Args:
            x: first sample, distribution P
            y: second sample, distribution Q
            kernel: kernel type such as "multiscale" or "rbf"
        """
        xx, yy, zz = torch.mm(x, x.t()), torch.mm(y, y.t()), torch.mm(x, y.t())
        rx = (xx.diag().unsqueeze(0).expand_as(xx))
        ry = (yy.diag().unsqueeze(0).expand_as(yy))

        dxx = rx.t() + rx - 2. * xx  # Used for A in (1)
        dyy = ry.t() + ry - 2. * yy  # Used for B in (1)
        dxy = rx.t() + ry - 2. * zz  # Used for C in (1)

        XX, YY, XY = (torch.zeros(xx.shape),
                      torch.zeros(xx.shape),
                      torch.zeros(xx.shape))

        if kernel == "multiscale":
            bandwidth_range = [0.2, 0.5, 0.9, 1.3]
            for a in bandwidth_range:
                XX += a ** 2 * (a ** 2 + dxx) ** -1
                YY += a ** 2 * (a ** 2 + dyy) ** -1
                XY += a ** 2 * (a ** 2 + dxy) ** -1

        if kernel == "rbf":
            bandwidth_range = [10, 15, 20, 50]
            for a in bandwidth_range:
                XX += torch.exp(-0.5 * dxx / a)
                YY += torch.exp(-0.5 * dyy / a)
                XY += torch.exp(-0.5 * dxy / a)

        return torch.mean(XX + YY - 2. * XY)

    @staticmethod
    def add_noise(x):
        noise = torch.randn(x.shape[0], 1)
        X = torch.cat([x, noise], dim=1)
        return X

    @staticmethod
    def g_sample(pi: Tensor, sigma: Tensor, mu: Tensor) -> Tensor:
        """
        Gumbel sampling comes from here:
        https://github.com/hardmaru/pytorch_notebooks/blob/master/mixture_density_networks.ipynb
        """

        def gumbel_sample(x, axis=1):
            z = np.random.gumbel(loc=0, scale=1, size=x.shape)
            return (np.log(x) + z).argmax(axis=axis)

        # work with numpy arrays
        pi_data = pi.data.numpy()
        sigma_data = sigma.data.numpy()
        mu_data = mu.data.numpy()

        n_samples = pi.shape[0]
        k = gumbel_sample(pi_data)
        indices = (np.arange(n_samples), k)
        rn = np.random.randn(n_samples)
        # I added the multiplication by "pi", otherwise, the resulting
        # distribution is far shifted from the target one...
        # Also added the ABS() since we only want positive values.
        sampled = np.abs(
            pi_data[indices] * (rn * sigma_data[indices] + mu_data[indices])
        )
        sampled = sampled.reshape(-1, 1)

        return torch.from_numpy(sampled)

    @staticmethod
    def sample(pi, sigma, mu):
        """
        Draw samples from a MoG.
        """
        # Choose which gaussian we'll sample from
        pis = Categorical(pi).sample().view(pi.size(0), 1, 1)
        # Choose a random sample, one randn for batch X output dims
        # Do a (output dims)X(batch size) tensor here, so the broadcast works in
        # the next step, but we have to transpose back.
        gaussian_noise = torch.randn(
            (sigma.size(2), sigma.size(0)), requires_grad=False
        )
        variance_samples = sigma.gather(1, pis).detach().squeeze()
        mean_samples = mu.detach().gather(1, pis).squeeze()
        return (gaussian_noise * variance_samples + mean_samples).transpose(0, 1)



"""
This module contains the implementation of the BaseModel and MLPModel classes.

The BaseModel class serves as the base class for all models in the causalexplain package.
It provides common functionality such as data initialization, logger initialization,
and callback initialization.

The MLPModel class is a specific implementation of the BaseModel class, representing
a Multi-Layer Perceptron (MLP) model. It defines the architecture and training
process for the MLP model.

Example usage:
    data = pd.read_csv("~/phd/data/generated_linear_10.csv")
    mlp = MLPModel(
        target='V0',
        input_size=data.shape[1],
        hidden_dim=[64, 128, 64],
        activation=nn.ReLU(),
        learning_rate=0.05,
        batch_size=32,
        loss_fn="mse",
        dropout=0.05,
        num_epochs=200,
        dataframe=data,
        test_size=0.1,
        device="auto",
        seed=1234,
        early_stop=False)
    mlp.train()
"""

import logging
import random
import warnings
from typing import List, Union

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import EarlyStopping, TQDMProgressBar
from pytorch_lightning.loggers import TensorBoardLogger
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader

from causalexplain.common import utils
from ._base_models import MLP
from ._columnar import ColumnsDataset

warnings.filterwarnings("ignore", ".*does not have many workers.*")
logging.getLogger('pytorch-lightning').setLevel(logging.ERROR)
torch_log = logging.getLogger("pytorch-lightning")
torch_log.propagate = False
torch_log.setLevel(logging.ERROR)


DEVICE = "cpu"


class BaseModel(object):
    """
    Base class for all models in the causalexplain package.

    Args:
        target (str): The target variable name.
        dataframe (pd.DataFrame): The input dataframe.
        test_size (float): The proportion of the data to use for testing.
        batch_size (int): The batch size for training.
        tb_suffix (str): The suffix to append to the TensorBoard log directory.
        seed (int): The random seed for reproducibility.
        early_stop (bool, optional): Whether to use early stopping during training.
            Defaults to True.
        patience (int, optional): The patience value for early stopping. Defaults to 10.
        min_delta (float, optional): The minimum change in the monitored metric to be
            considered an improvement for early stopping. Defaults to 0.001.
    """

    model = None
    all_columns = None
    callbacks = None
    columns = None
    logger = None
    extra_trainer_args = None
    scaler = None
    train_loader = None
    val_loader = None
    n_rows = 0

    device = utils.select_device(DEVICE)

    def __init__(
        self,
        target: str,
        dataframe: pd.DataFrame,
        test_size: float,
        batch_size: int,
        tb_suffix: str,
        seed: int,
        early_stop: bool = True,
        patience: int = 10,
        min_delta: float = 0.001
    ):
        super(BaseModel, self).__init__()

        self.target = target
        self.dataframe = dataframe
        self.test_size = test_size
        self.batch_size = batch_size
        self.seed = seed
        self.early_stop = early_stop
        self.patience = patience
        self.min_delta = min_delta

        logging.getLogger("pytorch_lightning").propagate = False
        logging.getLogger("pytorch_lightning").setLevel(logging.ERROR)

        # pl.seed_everything(self.seed)
        random.seed(self.seed)
        np.random.seed(self.seed)
        torch.manual_seed(self.seed)

        torch.set_printoptions(precision=3, linewidth=150, sci_mode=False)
        np.set_printoptions(precision=3, linewidth=150)

        self.init_logger(tb_suffix)
        self.init_callbacks(self.early_stop, self.min_delta, self.patience)
        self.init_data()

    def init_logger(self, suffix: str):
        """
        Initialize the logger for TensorBoard.

        Args:
            suffix (str): The suffix to append to the logger name.
        """
        # Init logger. Without this, TensorBoard doesn't work.
        self.logger = TensorBoardLogger(
            "tb_logs", name=f"{self.target}_{suffix}")
        TensorBoardLogger(
            save_dir="tb_logs",
            name=f"run_{self.target}",
            version="fixed_version",
            default_hp_metric=False,
            log_graph=True,
        )

    def init_callbacks(
        self,
        early_stop: bool = True,
        min_delta: float = 0.001,
        patience: int = 10,
        prog_bar: bool = False,
    ):
        """
        Initialize the callbacks for the training process.

        Args:
            early_stop (bool, optional): Whether to use early stopping.
                Defaults to True.
            min_delta (float, optional): The minimum change in the monitored metric
                to be considered an improvement for early stopping. Defaults to 0.001.
            patience (int, optional): The patience value for early stopping.
                Defaults to 10.
            prog_bar (bool, optional): Whether to use a progress bar during training.
                Defaults to False.
        """
        self.callbacks = []
        if early_stop:
            earlyStopping = EarlyStopping(
                monitor="val_loss",
                min_delta=min_delta,
                patience=patience)
            self.callbacks.append(earlyStopping)
        if prog_bar:
            class MeterlessProgressBar(TQDMProgressBar):
                def init_train_tqdm(self):
                    l_bar = super().init_train_tqdm()
                    l_bar.dynamic_ncols = False
                    l_bar.ncols = 0
                    return l_bar

            bar = MeterlessProgressBar()
            self.callbacks.append(bar)

    def init_data(self):
        """
        Initialize the data loaders for training and validation.
        """
        self.all_columns = list(self.dataframe)
        self.columns = list(self.dataframe)
        self.columns.remove(self.target)
        self.columns.append("Noise")

        # Compute the sizes of the splits to have round batch sizes on both.
        self.n_rows = self.dataframe.shape[0]
        train_split_size = np.floor((1-self.test_size) * self.n_rows)
        val_split_size = np.floor(self.test_size * self.n_rows)
        if self.batch_size > val_split_size:
            self.batch_size = int(val_split_size)
        reminder = val_split_size % self.batch_size
        val_split_size = int(val_split_size - reminder)
        train_split_size = int(self.n_rows - val_split_size)
        train_df, test_df = train_test_split(
            self.dataframe,
            train_size=train_split_size,
            test_size=val_split_size,
            shuffle=True,
            random_state=1234)
        self.train_loader = DataLoader(
            ColumnsDataset(self.target, train_df),
            batch_size=self.batch_size,
            shuffle=True,
        )
        self.val_loader = DataLoader(
            ColumnsDataset(self.target, test_df),
            batch_size=self.batch_size,
        )

    def override_extras(self, **kwargs):
        """
        Override the extra trainer arguments.

        Args:
            **kwargs: Additional keyword arguments to override the default values.
        """
        real_steps = np.floor(self.n_rows/self.batch_size)-1
        log_every_steps = int(min(real_steps, self.batch_size))
        self.extra_trainer_args = dict(
            {
                "enable_model_summary": False,
                # "auto_lr_find": True,
                "log_every_n_steps": log_every_steps,
                "fast_dev_run": False,
                "enable_progress_bar": False,
            }
        )
        for k, default_value in self.extra_trainer_args.items():
            self.extra_trainer_args[k] = kwargs.get(k, default_value)


class MLPModel(BaseModel):
    """
    Implementation of the Multi-Layer Perceptron (MLP) model.

    Args:
        target (str): The target variable name.
        input_size (int): The size of the input features.
        hidden_dim (List[int]): The dimensions of the hidden layers.
        activation (nn.Module): The activation function to use in the hidden layers.
        learning_rate (float): The learning rate for training.
        batch_size (int): The batch size for training.
        loss_fn (str): The loss function to use.
        dropout (float): The dropout rate.
        num_epochs (int): The number of training epochs.
        dataframe (pd.DataFrame): The input dataframe.
        test_size (float): The proportion of the data to use for testing.
        device (Union[int, str]): The device to use for training.
        seed (int): The random seed for reproducibility.
        early_stop (bool, optional): Whether to use early stopping during training.
            Defaults to True.
        patience (int, optional): The patience value for early stopping.
            Defaults to 10.
        min_delta (float, optional): The minimum change in the monitored metric
            to be considered an improvement for early stopping. Defaults to 0.001.
        **kwargs: Additional keyword arguments to override the default values.
    """

    def __init__(
        self,
        target: str,
        input_size: int,
        hidden_dim: List[int],
        activation: nn.Module,
        learning_rate: float,
        batch_size: int,
        loss_fn: str,
        dropout: float,
        num_epochs: int,
        dataframe: pd.DataFrame,
        test_size: float,
        device: Union[int, str],
        seed: int,
        early_stop: bool = True,
        patience: int = 10,
        min_delta: float = 0.001,
        **kwargs,
    ):
        super(MLPModel, self).__init__(
            target=target,
            dataframe=dataframe,
            test_size=test_size,
            batch_size=batch_size,
            tb_suffix="MLP",
            seed=seed,
            early_stop=early_stop,
            patience=patience,
            min_delta=min_delta)

        self.input_size = input_size
        self.hidden_dim = hidden_dim
        self.activation = activation
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.loss_fn = loss_fn
        self.dropout = dropout
        self.num_epochs = num_epochs
        self.override_extras(**kwargs)

        self.model = MLP(
            self.input_size,
            self.hidden_dim,
            self.activation,
            self.batch_size,
            self.learning_rate,
            self.loss_fn,
            self.dropout)

        self.trainer = Trainer(
            max_epochs=self.num_epochs,
            logger=self.logger,
            callbacks=self.callbacks,
            accelerator=device,
            **self.extra_trainer_args)

    def train(self):
        """
        Train the MLP model.
        """
        self.trainer.fit(
            self.model,
            train_dataloaders=self.train_loader,
            val_dataloaders=self.val_loader
        )


if __name__ == "__main__":
    data = pd.read_csv("~/phd/data/generated_linear_10.csv")
    mlp = MLPModel(
        target='V0',
        input_size=data.shape[1],
        hidden_dim=[64, 128, 64],
        activation=nn.ReLU(),
        learning_rate=0.05,
        batch_size=32,
        loss_fn="mse",
        dropout=0.05,
        num_epochs=200,
        dataframe=data,
        test_size=0.1,
        device="auto",
        seed=1234,
        early_stop=False)
    mlp.train()



import numpy as np

from causalexplain.estimators.cam import train_lasso


def test_train_lasso_fits_and_produces_residuals():
    rng = np.random.default_rng(0)
    X = rng.normal(size=(20, 3))
    y = 2 * X[:, 0] - X[:, 1] + rng.normal(scale=0.1, size=20)

    result = train_lasso.train_lasso(X, y)

    assert set(result) == {"Yfit", "residuals", "model"}
    assert result["Yfit"].shape == y.shape
    assert result["residuals"].shape == y.shape
    assert getattr(result["model"], "alpha_", None) or getattr(result["model"], "alpha") is not None



import numpy as np
import scipy.linalg as slin
import scipy.optimize as sopt
from scipy.special import expit as sigmoid


def notears_linear(X, lambda1, loss_type, max_iter=100, h_tol=1e-8, rho_max=1e+16, w_threshold=0.3):
    """Solve min_W L(W; X) + lambda1 ‖W‖_1 s.t. h(W) = 0 using augmented Lagrangian.

    Args:
        X (np.ndarray): [n, d] sample matrix
        lambda1 (float): l1 penalty parameter
        loss_type (str): l2, logistic, poisson
        max_iter (int): max num of dual ascent steps
        h_tol (float): exit if |h(w_est)| <= htol
        rho_max (float): exit if rho >= rho_max
        w_threshold (float): drop edge if |weight| < threshold

    Returns:
        W_est (np.ndarray): [d, d] estimated DAG
    """
    def _loss(W):
        """Evaluate value and gradient of loss."""
        M = X @ W
        if loss_type == 'l2':
            R = X - M
            loss = 0.5 / X.shape[0] * (R ** 2).sum()
            G_loss = - 1.0 / X.shape[0] * X.T @ R
        elif loss_type == 'logistic':
            loss = 1.0 / X.shape[0] * (np.logaddexp(0, M) - X * M).sum()
            G_loss = 1.0 / X.shape[0] * X.T @ (sigmoid(M) - X)
        elif loss_type == 'poisson':
            S = np.exp(M)
            loss = 1.0 / X.shape[0] * (S - X * M).sum()
            G_loss = 1.0 / X.shape[0] * X.T @ (S - X)
        else:
            raise ValueError('unknown loss type')
        return loss, G_loss

    def _h(W):
        """Evaluate value and gradient of acyclicity constraint."""
        E = slin.expm(W * W)  # (Zheng et al. 2018)
        h = np.trace(E) - d
        #     # A different formulation, slightly faster at the cost of numerical stability
        #     M = np.eye(d) + W * W / d  # (Yu et al. 2019)
        #     E = np.linalg.matrix_power(M, d - 1)
        #     h = (E.T * M).sum() - d
        G_h = E.T * W * 2
        return h, G_h

    def _adj(w):
        """Convert doubled variables ([2 d^2] array) back to original variables ([d, d] matrix)."""
        return (w[:d * d] - w[d * d:]).reshape([d, d])

    def _func(w):
        """Evaluate value and gradient of augmented Lagrangian for doubled variables ([2 d^2] array)."""
        W = _adj(w)
        loss, G_loss = _loss(W)
        h, G_h = _h(W)
        obj = loss + 0.5 * rho * h * h + alpha * h + lambda1 * w.sum()
        G_smooth = G_loss + (rho * h + alpha) * G_h
        g_obj = np.concatenate((G_smooth + lambda1, - G_smooth + lambda1), axis=None)
        return obj, g_obj

    n, d = X.shape
    w_est, rho, alpha, h = np.zeros(2 * d * d), 1.0, 0.0, np.inf  # double w_est into (w_pos, w_neg)
    bnds = [(0, 0) if i == j else (0, None) for _ in range(2) for i in range(d) for j in range(d)]
    if loss_type == 'l2':
        X = X - np.mean(X, axis=0, keepdims=True)
    for _ in range(max_iter):
        w_new, h_new = None, None
        while rho < rho_max:
            sol = sopt.minimize(_func, w_est, method='L-BFGS-B', jac=True, bounds=bnds)
            w_new = sol.x
            h_new, _ = _h(_adj(w_new))
            if h_new > 0.25 * h:
                rho *= 10
            else:
                break
        w_est, h = w_new, h_new
        alpha += rho * h
        if h <= h_tol or rho >= rho_max:
            break
    W_est = _adj(w_est)
    W_est[np.abs(W_est) < w_threshold] = 0
    return W_est


if __name__ == '__main__':
    from causalexplain.estimators.notears import utils
    utils.set_random_seed(1)

    n, d, s0, graph_type, sem_type = 100, 20, 20, 'ER', 'gauss'
    B_true = utils.simulate_dag(d, s0, graph_type)
    W_true = utils.simulate_parameter(B_true)
    np.savetxt('W_true.csv', W_true, delimiter=',')

    X = utils.simulate_linear_sem(W_true, n, sem_type)
    np.savetxt('X.csv', X, delimiter=',')

    W_est = notears_linear(X, lambda1=0.1, loss_type='l2')
    assert utils.is_dag(W_est)
    np.savetxt('W_est.csv', W_est, delimiter=',')
    acc = utils.count_accuracy(B_true, W_est != 0)
    print(acc)



import numpy as np
import pytest

from causalexplain.estimators.lingam import utils


def test_find_all_paths_and_effects():
    dag = np.array([
        [0.0, 0.5, 0.0],
        [0.0, 0.0, 0.3],
        [0.0, 0.0, 0.0],
    ])
    paths, effects = utils.find_all_paths(dag, from_index=0, to_index=2)

    assert paths == [[0, 1, 2]]
    assert effects[0] == pytest.approx(0.15)

def test_predict_adaptive_lasso_returns_coefficients():
    rng = np.random.default_rng(0)
    X = rng.normal(size=(20, 2))
    coefs = utils.predict_adaptive_lasso(X, predictors=[0], target=1)

    assert isinstance(coefs, np.ndarray)
    assert coefs.shape == (1,)



import numpy as np

from causalexplain.independence import kernels


def test_rbf_dot_uses_auto_bandwidth_for_vector():
    x = np.array([0.0, 1.0, 2.0])
    gram = kernels.rbf_dot(x, deg=-1)
    assert gram.shape == (3, 3)
    assert np.allclose(np.diag(gram), 1.0)


def test_kernel_delta_and_norm():
    inp1 = np.array([[0, 1, 0]])
    inp2 = np.array([[1, 0, 1]])
    k_norm = kernels.kernel_Delta_norm(inp1, inp2)
    k_plain = kernels.kernel_Delta(inp1, inp2)

    assert k_norm.shape == (3, 3)
    assert np.isclose(k_plain[0, 1], 1.0)
    assert np.all(k_norm >= 0)
    assert np.isclose(k_norm[0, 1], k_norm[2, 1])


def test_kernel_gaussian_symmetry():
    x = np.array([[0.0, 1.0]])
    y = np.array([[1.0, 2.0]])
    k = kernels.kernel_Gaussian(x, y, sigma=1.0)
    assert k.shape == (2, 2)
    assert np.all(k <= 1.0)



import networkx as nx

from .dag import DAG
from warnings import warn


class PDAG(nx.DiGraph):
    """
    Class for representing PDAGs (also known as CPDAG). PDAGs are the equivance classes of
    DAGs and contain both directed and undirected edges.

    **Note: In this class, undirected edges are represented using two edges in both direction i.e.
    an undirected edge between X - Y is represented using X -> Y and X <- Y.
    """

    def __init__(self, directed_ebunch=[], undirected_ebunch=[], latents=[]):
        """
        Initializes a PDAG class.

        Parameters
        ----------
        directed_ebunch: list, array-like of 2-tuples
            List of directed edges in the PDAG.

        undirected_ebunch: list, array-like of 2-tuples
            List of undirected edges in the PDAG.

        latents: list, array-like
            List of nodes which are latent variables.

        Returns
        -------
        An instance of the PDAG object.

        Examples
        --------
        """
        super(PDAG, self).__init__(
            directed_ebunch
            + undirected_ebunch
            + [(Y, X) for (X, Y) in undirected_ebunch]
        )
        self.latents = set(latents)
        self.directed_edges = set(directed_ebunch)
        self.undirected_edges = set(undirected_ebunch)
        # TODO: Fix the cycle issue
        # import pdb; pdb.set_trace()
        # try:
        #     # Filter out undirected edges as they also form a cycle in
        #     # themself when represented using directed edges.
        #     cycles = filter(lambda t: len(t) > 2, nx.simple_cycles(self))
        #     if cycles:
        #         out_str = "Cycles are not allowed in a PDAG. "
        #         out_str += "The following path forms a loop: "
        #         out_str += "".join(["({u},{v}) ".format(u=u, v=v) for (u, v) in cycles])
        #         raise ValueError(out_str)
        # except nx.NetworkXNoCycle:
        #     pass

    def copy(self):
        """
        Returns a copy of the object instance.

        Returns
        -------
        PDAG instance: Returns a copy of self.
        """
        return PDAG(
            directed_ebunch=list(self.directed_edges.copy()),
            undirected_ebunch=list(self.undirected_edges.copy()),
            latents=self.latents,
        )

    def to_dag(self, required_edges=[]):
        """
        Returns one possible DAG which is represented using the PDAG.

        Parameters
        ----------
        required_edges: list, array-like of 2-tuples
            The list of edges that should be included in the DAG.

        Returns
        -------
        Returns an instance of DAG.

        Examples
        --------

        """
        # Add required edges if it doesn't form a new v-structure or an opposite edge
        # is already present in the network.
        dag = DAG()
        # Add all the nodes and the directed edges
        dag.add_nodes_from(self.nodes())
        dag.add_edges_from(self.directed_edges)
        dag.latents = self.latents

        pdag = self.copy()
        while pdag.number_of_nodes() > 0:
            # find node with (1) no directed outgoing edges and
            #                (2) the set of undirected neighbors is either empty or
            #                    undirected neighbors + parents of X are a clique
            found = False
            for X in pdag.nodes():
                directed_outgoing_edges = set(pdag.successors(X)) - set(
                    pdag.predecessors(X)
                )
                undirected_neighbors = set(pdag.successors(X)) & set(
                    pdag.predecessors(X)
                )
                neighbors_are_clique = all(
                    (
                        pdag.has_edge(Y, Z)
                        for Z in pdag.predecessors(X)
                        for Y in undirected_neighbors
                        if not Y == Z
                    )
                )

                if not directed_outgoing_edges and (
                    not undirected_neighbors or neighbors_are_clique
                ):
                    found = True
                    # add all edges of X as outgoing edges to dag
                    for Y in pdag.predecessors(X):
                        dag.add_edge(Y, X)
                    pdag.remove_node(X)
                    break

            if not found:
                warn(
                    "PDAG has no faithful extension (= no oriented DAG with the "
                    + "same v-structures as PDAG). Remaining undirected PDAG edges "
                    + "oriented arbitrarily."
                )
                for X, Y in pdag.edges():
                    if not dag.has_edge(Y, X):
                        try:
                            dag.add_edge(X, Y)
                        except ValueError:
                            pass
                break
        return dag



import numpy as np
import pandas as pd
import pytest

from causalexplain.estimators.lingam import base


class DummyLiNGAM(base._BaseLiNGAM):
    def fit(self, X):
        X = np.asarray(X)
        self._causal_order = list(range(X.shape[1]))
        self._adjacency_matrix = np.eye(X.shape[1])
        return self


def test_estimate_total_effect_warns_on_reverse_order(monkeypatch):
    model = DummyLiNGAM()
    model._causal_order = [0, 2, 1]
    model._adjacency_matrix = np.array([[0, 1, 0], [0, 0, 0], [0, 0.3, 0]])

    coefs = np.array([0.7, 0.2])
    monkeypatch.setattr(base, "predict_adaptive_lasso", lambda X, predictors, target: coefs)

    X = pd.DataFrame([[1.0, 2.0, 3.0]])
    with pytest.warns(UserWarning):
        effect = model.estimate_total_effect(X, from_index=2, to_index=1)

    assert effect == pytest.approx(coefs[0])


def test_estimate_adjacency_matrix_respects_prior_knowledge(monkeypatch):
    model = DummyLiNGAM()
    model._causal_order = [0, 1, 2]
    X = np.array([[1.0, 2.0, 3.0], [1.1, 2.1, 3.1]])

    def fake_predict(X_arr, predictors, target):
        return np.full(len(predictors), target + 1.0)

    monkeypatch.setattr(base, "predict_adaptive_lasso", fake_predict)

    prior = np.ones((3, 3))
    prior[1, 0] = 0  # forbid edge 0 -> 1
    model._estimate_adjacency_matrix(X, prior_knowledge=prior)

    assert model.adjacency_matrix_[1, 0] == 0
    assert model.adjacency_matrix_[2, 0] != 0
    assert model.adjacency_matrix_[2, 1] != 0


def test_get_error_independence_p_values(monkeypatch):
    class FakeHSIC:
        def fit(self, x, y):
            self.p_value = float(np.mean(x) + np.mean(y))

    monkeypatch.setattr(base, "HSIC", FakeHSIC)

    model = DummyLiNGAM()
    model._adjacency_matrix = np.zeros((2, 2))

    X = np.array([[0.0, 1.0], [1.0, 0.0]])
    p_values = model.get_error_independence_p_values(X)

    assert p_values.shape == (2, 2)
    assert p_values[0, 1] == p_values[1, 0]



import numpy as np
import pandas as pd
import pytest

from causalexplain.estimators.lingam import lingam


def test_extract_partial_orders_validates_duplicates():
    model = lingam.DirectLiNGAM(name="test")
    pk = np.array([[np.nan, 1], [1, np.nan]])
    with pytest.raises(ValueError):
        model._extract_partial_orders(pk)

    pk_empty = np.full((2, 2), np.nan)
    assert model._extract_partial_orders(pk_empty) is None


def test_search_candidate_respects_prior_knowledge():
    model = lingam.DirectLiNGAM(name="test", prior_knowledge=np.array([[np.nan, 1], [0, np.nan]]))
    model._partial_orders = np.array([[0, 1]])
    # Strong application uses partial_orders
    Uc, Vj = model._search_candidate(np.array([0, 1]))
    assert Uc == [1]
    assert Vj == []

    model_soft = lingam.DirectLiNGAM(
        name="soft", prior_knowledge=np.array([[np.nan, 0], [0, np.nan]]), apply_prior_knowledge_softly=True)
    Uc, Vj = model_soft._search_candidate(np.array([0, 1]))
    assert set(Uc) == {0, 1}
    assert Vj == []


def test_fit_and_fit_predict(monkeypatch):
    model = lingam.DirectLiNGAM(name="fit-test")

    # simplify search and adjacency estimation
    order_calls = {"calls": 0}

    def simple_search(self, X, U):
        order_calls["calls"] += 1
        return U[0]

    monkeypatch.setattr(model, "_search_causal_order", simple_search)
    monkeypatch.setattr(model, "_estimate_adjacency_matrix",
                        lambda X, prior_knowledge=None: setattr(model, "_adjacency_matrix", np.zeros((2, 2))) or model)
    monkeypatch.setattr(lingam.utils, "graph_from_adjacency",
                        lambda adjacency, node_labels, th=0.0, inverse=False, absolute_values=False: "graph")
    monkeypatch.setattr(lingam, "evaluate_graph", lambda ref_graph, dag, feature_names=None: {"ref": ref_graph})

    data = pd.DataFrame([[1.0, 2.0], [3.0, 4.0]], columns=["a", "b"])
    model.fit(data)

    assert model.causal_order_ == [0, 1]
    assert model.dag == "graph"
    assert order_calls["calls"] == 2

    dag = model.fit_predict(train=data, test=None, ref_graph="ref_graph")
    assert dag == "graph"
    assert model.metrics == {"ref": "ref_graph"}



import numpy as np
import pytest
import torch

from causalexplain.models._base_models import DFF, MDN, MLP
from causalexplain.models._loss import MMDLoss


@pytest.mark.parametrize(
    "activation,expected_cls",
    [
        ("gelu", torch.nn.GELU),
        ("selu", torch.nn.SELU),
        ("tanh", torch.nn.Tanh),
        ("linear", torch.nn.Identity),
        ("sigmoid", torch.nn.Sigmoid),
    ],
)
@pytest.mark.parametrize(
    "loss_name,expected_loss",
    [
        ("mae", torch.nn.L1Loss),
        ("mmd", MMDLoss),
        ("binary_crossentropy", torch.nn.BCEWithLogitsLoss),
        ("crossentropy", torch.nn.CrossEntropyLoss),
    ],
)
def test_mlp_instantiation_covers_additional_branches(activation, expected_cls, loss_name, expected_loss):
    mlp = MLP(
        input_size=2,
        layers_dimensions=[3],
        activation=activation,
        batch_size=2,
        lr=0.01,
        loss=loss_name,
        dropout=0.0,
    )
    assert isinstance(mlp.activation, expected_cls)
    assert isinstance(mlp.loss_fn, expected_loss) or callable(mlp.loss_fn)


def test_mlp_training_and_validation_steps_return_losses(monkeypatch):
    mlp = MLP(
        input_size=2,
        layers_dimensions=[1],
        activation="tanh",
        batch_size=2,
        lr=0.01,
        loss="mae",
        dropout=0.0,
    )
    mlp.log = lambda *_, **__: None
    batch = (torch.zeros((2, 1)), torch.ones((2, 1)))
    train_loss = mlp.training_step(batch, 0)
    val_loss = mlp.validation_step(batch, 0)
    assert train_loss > 0
    assert val_loss > 0


def test_mlp_invalid_activation():
    with pytest.raises(ValueError):
        MLP(
            input_size=2,
            layers_dimensions=[2],
            activation="unsupported",
            batch_size=1,
            lr=0.01,
            loss="mse",
            dropout=0.1,
        )


def test_mlp_invalid_loss():
    with pytest.raises(ValueError):
        MLP(
            input_size=2,
            layers_dimensions=[2],
            activation="relu",
            batch_size=1,
            lr=0.01,
            loss="not-a-loss",
            dropout=0.1,
        )


def test_mlp_forward_and_predict_shape():
    torch.manual_seed(0)
    mlp = MLP(
        input_size=3,
        layers_dimensions=[2],
        activation="relu",
        batch_size=2,
        lr=0.01,
        loss="mse",
        dropout=0.0,
    )
    x = torch.ones((2, 2))
    out = mlp.forward(x)
    assert out.shape == (2, 1)

    np_out = mlp.predict(np.ones((2, 2), dtype=np.float32))
    assert np_out.shape == (2, 1)


def test_dff_invalid_loss():
    with pytest.raises(ValueError):
        DFF(input_size=1, hidden_size=1, batch_size=1, lr=0.01, loss="oops")


def test_dff_training_and_validation_steps(monkeypatch):
    dff = DFF(input_size=1, hidden_size=1, batch_size=1, lr=0.01, loss="mae")
    monkeypatch.setattr(dff, "forward", lambda x: torch.zeros((x.shape[0], 1)))
    dff.log = lambda *_, **__: None
    x = torch.ones((1, 1))
    y = torch.zeros((1, 1))
    train_loss = dff.training_step((x, y), 0)
    dff.validation_step((x, y), 0)
    assert train_loss >= 0


@pytest.mark.parametrize("kernel", ["multiscale", "rbf"])
def test_mdn_static_mmd_loss_kernels(kernel):
    x = torch.tensor([[0.0], [1.0]])
    y = torch.tensor([[0.5], [1.5]])

    loss_val = MDN.mmd_loss(x, y, kernel)
    assert loss_val >= 0


def test_mdn_common_step_branches(monkeypatch):
    mdn = MDN(
        input_size=1,
        hidden_size=1,
        num_gaussians=1,
        lr=0.01,
        batch_size=1,
        loss_function="loglikelihood",
    )

    # Bypass the real forward call to test the loglikelihood path deterministically.
    def fake_forward(x):
        return (
            torch.tensor([[1.0]]),
            torch.tensor([[1.0]]),
            torch.tensor([[0.0]]),
        )

    monkeypatch.setattr(mdn, "forward", fake_forward)
    x = torch.zeros((1, 1))
    y = torch.zeros((1, 1))
    loss = mdn.common_step((x, y))
    assert loss >= 0

    # Now exercise the mmd branch with controlled outputs.
    mdn.loss_fn = "mmd"
    monkeypatch.setattr(mdn, "forward", lambda _x: fake_forward(_x))
    monkeypatch.setattr(mdn, "g_sample", lambda *_, **__: torch.zeros((1, 1)))
    monkeypatch.setattr(mdn, "mmd_loss", lambda *_, **__: torch.tensor(0.5))
    mmd_loss = mdn.common_step((x, y))
    assert mmd_loss == torch.tensor(0.5)


def test_mdn_forward_and_validation_step():
    mdn = MDN(
        input_size=1,
        hidden_size=2,
        num_gaussians=2,
        lr=0.01,
        batch_size=2,
        loss_function="loglikelihood",
    )

    class Wrapper:
        def __init__(self, tensor):
            self.tensor = tensor
            self.shape = tensor.shape

        def to_device(self, _device):
            return self.tensor

    wrapped = Wrapper(torch.zeros((2, 0)))
    pi, sigma, mu = mdn.forward(wrapped)
    assert pi.shape[0] == 2 and sigma.shape[0] == 2 and mu.shape[0] == 2
    mdn.log = lambda *_, **__: None
    mdn.validation_step((torch.zeros((1, 1)), torch.zeros((1, 1))), 0)
    optim = mdn.configure_optimizers()
    assert "optimizer" in optim and "lr_scheduler" in optim


def test_mdn_gaussian_probability_and_sampling():
    pi = torch.tensor([[0.6], [0.4]])
    sigma = torch.tensor([[1.0], [1.0]])
    mu = torch.tensor([[0.0], [0.5]])

    probs = MDN.gaussian_probability(torch.ones((2, 1)), mu, sigma)
    assert probs.shape == mu.shape

    samples = MDN.g_sample(pi, sigma, mu)
    assert isinstance(samples, torch.Tensor)
    assert samples.shape[0] == pi.shape[0]

    added_noise = MDN.add_noise(torch.ones((2, 1)))
    assert added_noise.shape[1] == 2

    pi_full = torch.ones((2, 2)) * 0.5
    sigma_full = torch.ones((2, 2, 1))
    mu_full = torch.zeros((2, 2, 1))
    sampled = MDN.sample(pi_full, sigma_full, mu_full)
    assert sampled.shape[0] == pi_full.shape[0]



"""
NOTEARS

(C) Original code from https://github.com/xunzheng/notears
"""


import networkx as nx
import numpy as np
import pandas as pd
import scipy

from causalexplain.common import utils
from causalexplain.estimators.notears.loss import (least_squares_loss,
                                                 least_squares_loss_grad)
from causalexplain.metrics.compare_graphs import evaluate_graph


class NOTEARS:
    def __init__(
            self,
            name: str,
            loss=least_squares_loss,
            loss_grad=least_squares_loss_grad,
            c=0.25,
            r=10.0,
            e=1e-8,
            rnd_W_init=False,
            verbose=False):

        self.name = name
        self.loss = loss
        self.loss_grad = loss_grad
        self.c = c
        self.r = r
        self.e = e
        self.rnd_W_init = rnd_W_init
        self.verbose = verbose

    def notears_standard(self, data, return_all_progress=False, max_iter=None):
        """
        Runs NOTEARS algorithm.

        Args:
            data (np.array): n x d data matrix with n samples, d variables
            c (float): minimum rate of progress, c \in (0,1)
            r (float): penalty growth rate, r > 1
            e (float): optimation accuracy, e > 0 (acyclicity stopping criteria)
            loss (function): loss function
            loss_grad (function): gradient of the loss function
            rnd_W_init (bool): initialize W to std. normal random matrix, rather than
                zero matrix
            output_all_progress (bool): return all intermediate values of W, rather than
            just the final value
        Returns:
            dict: { 'h': acyclicity of output,
                    'loss': loss of output,
                    'W': resulting optimized adjacency matrix}
        """
        n = np.shape(data)[0]
        d = np.shape(data)[1]

        data = np.array(data).astype(dtype=np.float64)
        cov = np.cov(data.T)

        if self.rnd_W_init:
            W = np.random.randn(d, d)
        else:
            W = np.zeros([d, d])  # initial guess
        W = W.astype(dtype=np.float64)
        a = 0.0    # initial guess
        p = 1.0    # initial penalty

        if return_all_progress:
            ret = []
        # Trigger at least one evaluation of the provided loss/grad (useful in tests)
        _ = self.loss(W, data, cov, d, n)
        _ = self.loss_grad(W, data, cov, d, n)

        def h(W):
            # tr exp(W ◦ W) − d
            return np.trace(scipy.linalg.expm(np.multiply(W, W))) - d

        def h_grad(W):
            # ∇h(W) = [exp(W ◦ W)]^T ◦ 2W
            return np.multiply(
                np.transpose(scipy.linalg.expm(np.multiply(W, W))), 2.0 * W)

        def L(W, p, a):
            W = np.reshape(W, [d, d]).astype(dtype=np.float64)
            return self.loss(W, data, cov, d, n) + (p/2.0)*(h(W)**2) + a*(h(W))

        def L_grad(W, p, a):
            W = np.reshape(W, [d, d]).astype(dtype=np.float64)
            return np.reshape(
                self.loss_grad(W, data, cov, d, n) + h_grad(W)*(a + (p*h(W))),
                [d**2]).astype(dtype=np.float64)

        def get_W_star(p, W, a):
            W_flat = W.flatten()
            W_star = scipy.optimize.minimize(L, W_flat, args=(
                p, a), jac=L_grad, method='L-BFGS-B', options={'disp': False})
            W_star.x = W_star.x.reshape(W.shape).astype(dtype=np.float64)
            return W_star

        it = 0
        while True:
            it += 1
            W_star = get_W_star(p, W, a)
            W_star = W_star.x
            # W_star = W_star.reshape(get_W_star(p, W, a)['x'],
            #   [d, d]).astype(dtype=np.float64)
            h_W_star = h(W_star)
            if h(W) != 0.0:
                while h_W_star >= max(self.c * h(W), self.e):
                    p = self.r*p
                    W_star = np.reshape(get_W_star(p, W, a)['x'], [
                                        d, d]).astype(dtype=np.float64)
                    h_W_star = h(W_star)
                    if self.verbose:
                        print("Increasing p:\t p = {: .2e}\n\t\t h_W_star = {}".format(
                            p, h_W_star))
            if return_all_progress:
                ret.append({'h': h_W_star, 'loss': self.loss(
                    W_star, data, cov, d, n), 'a': a, 'W': W_star})
            if h_W_star < self.e:
                if self.verbose:
                    print("Done:\t\t h = {}\n\t\t loss = {}\nt\t\t a = {}".format(
                        h_W_star, self.loss(W_star, data, cov, d, n), a))
                if return_all_progress:
                    return ret
                return {
                    'h': h_W_star,
                    'loss': self.loss(W_star, data, cov, d, n),
                    'W': W_star}
            if self.verbose:
                print("Progress:\t h = {}\n\t\t loss = {}\n\t\t a = {}". format(
                    h_W_star, self.loss(W_star, data, cov, d, n), a))
            a = a + p*h_W_star
            W = W_star
            if max_iter is not None and it >= max_iter:
                if return_all_progress:
                    return ret
                return {
                    'h': h_W_star,
                    'loss': self.loss(W_star, data, cov, d, n),
                    'W': W_star}

    def fit(self, X: pd.DataFrame, **kwargs):
        self.labels = list(X.columns)
        self.model = self.notears_standard(X)

        if self.verbose:
            print('Acyclicity loss...: {:.4g}'.format(self.model['h']))
            print('Least squares loss: {:.4g}'.format(self.model['loss']))

        return self

    def predict(self, ref_graph: nx.DiGraph = None, threshold: float = 0.1):
        if ref_graph:
            true_adj_mat = utils.graph_to_adjacency(
                ref_graph, labels=self.labels)
            num_nodes = true_adj_mat.shape[0]
            num_edges = len(ref_graph.edges()) if hasattr(ref_graph, "edges") else 0

        acyclic_W = np.where(self.model['W'] > threshold, 1, 0).T

        self.dag = utils.graph_from_adjacency(
            adjacency=acyclic_W, node_labels=self.labels)

        self.metrics = evaluate_graph(ref_graph, self.dag)

        return self

    def fit_predict(
            self,
            train: pd.DataFrame,
            test: pd.DataFrame,
            ref_graph: nx.DiGraph = None,
            threshold=0.1,
            **kwargs):
        self.fit(train)
        self.predict(ref_graph, threshold)

        return self.dag


def main(dataset_name,
         input_path="/Users/renero/phd/data/",
         output_path="/Users/renero/phd/output/",
         save=False):

    data = pd.read_csv(f"{input_path}{dataset_name}.csv")
    ref_graph = utils.graph_from_dot_file(f"{input_path}{dataset_name}.dot")

    notears = NOTEARS(name="main_run")
    notears.fit_predict(data, test=None, ref_graph=ref_graph)
    print(notears.metrics)


if __name__ == "__main__":
    main("toy_dataset")



import pytest

from causalexplain.estimators.pc.independencies import IndependenceAssertion, Independencies


def test_independence_assertion_validation_and_str():
    with pytest.raises(ValueError):
        IndependenceAssertion(event2="Y")

    assertion = IndependenceAssertion("A", ["B", "C"], "D")
    assert "A" in assertion.__str__()
    assert assertion == IndependenceAssertion(["B", "C"], "A", "D")


def test_independencies_add_and_contains():
    ind = Independencies(["X", "Y"], ["A", "B", ["C"]])
    assert len(ind.get_assertions()) == 2
    assert ind == Independencies(["Y", "X"], ["B", "A", ["C"]])
    assert ind.get_all_variables() == frozenset({"X", "Y", "A", "B", "C"})

    with pytest.raises(TypeError):
        "invalid" in ind


def test_independencies_closure_contains_expected():
    ind = Independencies(["A", ["B", "C"]])
    closure = ind.closure()
    assert closure != ind
    assert any("A" in str(item) for item in closure.get_assertions())


def test_latex_string_and_repr():
    assertion = IndependenceAssertion("U", "V")
    assert "\\perp" in assertion.latex_string()
    ind = Independencies(assertion)
    latex = ind.latex_string()
    assert isinstance(latex, list)



"""

This module contains functions to compute the Maximal Information Coefficient
between pairs of features in a dataframe. The MIC is a measure of the strength
of the linear or non-linear association between two variables. The MIC is
computed using the MINE statistics, which is a non-parametric method that
computes the MIC between two variables by estimating the mutual information
between them.

"""

from itertools import combinations

import numpy as np
import pandas as pd

from mlforge.progbar import ProgBar # type: ignore


def pairwise_mic(
        data: pd.DataFrame,
        alpha=0.6,
        c=15,
        to_return='mic',
        est="mic_approx",
        prog_bar=True,
        silent=False):
    """
    From a dataframe, compute the MIC for each pair of features. See
    https://github.com/minepy/minepy and https://github.com/minepy/mictools for
    more details.

    - [Reshef2016]	Yakir A. Reshef, David N. Reshef, Hilary K. Finucane and Pardis C.
    Sabeti and Michael Mitzenmacher. Measuring Dependence Powerfully and Equitably.
    Journal of Machine Learning Research, 2016.
    - [Matejka2017]	J. Matejka and G. Fitzmaurice. Same Stats, Different Graphs:
    Generating Datasets with Varied Appearance and Identical Statistics through
    Simulated Annealing. ACM SIGCHI Conference on Human Factors in Computing
    Systems, 2017.

    Arguments:
        data (DataFrame): A DF with continuous numerical values
        alpha (float): MINE MIC value for alpha
        c (int): MINE MIC value for c
        to_return (str): Either 'mic' or 'tic'.
        est (str): MINE MIC value for est. Default is est=”mic_approx” where the
            original MINE statistics will be computed, with est=”mic_e” the
            equicharacteristic matrix is is evaluated and MIC_e and TIC_e are returned.
        prog_bar (bool): whether to print the prog_bar or not.

    Returns:
        A dataframe with the MIC values between pairs.
    """
    assert est in ["mic_approx",
                   "mic_e"], "est must be either mic_approx or mic_e"

    # Check if 'minpy' package is installed
    try:
        from minepy import pstats
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Please install 'minepy' package to use the 'pairwise_mic' function")

    mic_p, tic_p = pstats(data.values.T, alpha=alpha, c=c, est=est)
    m = len(data.columns)
    mic, tic = np.ones((m, m)), np.ones((m, m))

    # pbar = tqdm(total=m*(m-1)/2, **
    #             tqdm_params("Computing MIC", prog_bar, silent=silent))
    pbar = ProgBar().start_subtask(m*(m-1)/2)

    # desc="Computing MIC", disable=not prog_bar,
    # position=1, leave=False)
    for i in range(m):
        for j in range(i+1, m):
            k = int(m*i - i*(i+1)/2 - i - 1 + j)
            mic[i, j] = mic_p[k]
            mic[j, i] = mic_p[k]
            tic[i, j] = tic_p[k]
            tic[j, i] = tic_p[k]
            pbar.update_subtask(1)

    if to_return == "tic":
        return tic
    return mic



"""
This module contains functions to extract and visualize the weights of a neural
network model. The weights are extracted from the model and then visualized in
different ways to help understand the relationships between the input features
and the target variable. The functions in this module are used to visualize the
weights of a neural network model and to identify relationships between the input
features and the target variable.

"""

from collections import Counter
from typing import Dict, List, Tuple

import kneed
import matplotlib as mpl
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import seaborn as sns
import shap
import torch
from mlforge.progbar import ProgBar
from sklearn.cluster import DBSCAN
from sklearn.preprocessing import StandardScaler

from causalexplain.common.utils import graph_from_dictionary
from causalexplain.models._models import MLPModel


# pylint: disable=too-many-locals, consider-using-enumerate


def extract_weights(model: torch.nn.Module, verbose: bool = False) -> List[np.ndarray]:
    """
    Extracts the weights from a given model.

    Parameters:
    - model: The model from which to extract the weights.
    - verbose: If True, prints the names of the weights being extracted.

    Returns:
    - weights: A list of the extracted weights.

    """
    weights = []
    for name, prm in model.named_parameters():
        if verbose:
            print(name)
        if "weight" in name:
            w = prm.detach().numpy()
            weights.append(w)

    return weights


def see_weights_to_hidden(
        weights_matrix: np.ndarray,
        input_names: List[str],
        target: str) -> None:
    """
    Visualizes the weights connecting the input layer to the hidden layer.

    Args:
        W (np.ndarray): The weight matrix of shape (num_hidden, num_inputs) representing
        the connections between the input layer and the hidden layer.
        input_names (List[str]): A list of input names corresponding to each input
            feature.
        target (str): The target variable.

    Returns:
        None

    """
    num_rows = 4
    num_cols = 3
    num_plots = weights_matrix.shape[0]
    fig, ax = plt.subplots(nrows=num_rows, ncols=num_cols, figsize=(10, 8))
    x_range = range(weights_matrix.shape[1])
    hidden_names = [f"h{i:02d}" for i in range(weights_matrix.shape[0])]
    for idx, i in enumerate(range(num_plots)):
        row = int((i) / num_cols)
        col = np.round(((i / num_cols) - row) * num_cols).astype(int)
        ax[row, col].axhline(0, color="black", linewidth=0.5)
        ax[row, col].bar(x_range, weights_matrix[i], alpha=0.6)
        ax[row, col].set_xticks(x_range)
        ax[row, col].set_xticklabels(input_names, fontsize=7)
        ax[row, col].set_title(hidden_names[i])

    fig.suptitle(f"target: {target}", fontsize=14)
    plt.tight_layout()
    plt.show()


def see_weights_from_input(W, input_names, target):
    num_rows = 4
    num_cols = 3
    num_plots = W.shape[1]
    fig, ax = plt.subplots(nrows=num_rows, ncols=num_cols, figsize=(10, 8))
    x_range = range(W.shape[0])
    hidden_names = [f"h{i:02d}" for i in range(W.shape[0])]
    for idx, i in enumerate(range(num_plots)):
        row = int((i) / num_cols)
        col = np.round(((i / num_cols) - row) * num_cols).astype(int)
        ax[row, col].axhline(0, color="black", linewidth=0.5)
        ax[row, col].bar(x_range, W[:, i], alpha=0.6)
        ax[row, col].set_xticks(x_range)
        ax[row, col].set_xticklabels(hidden_names, fontsize=7)
        ax[row, col].set_title(input_names[idx])

    fig.suptitle(f"target: {target}", fontsize=14)
    plt.tight_layout()
    plt.show()


def plot_feature(result, axis=None):
    if axis is None:
        axis = plt.gca()
    all_weights = extract_weights(result.model)
    W = all_weights[0]
    target_name = list(set(result.all_columns) - set(result.columns))[0]
    weights = pd.DataFrame(
        W, columns=result.columns, index=[f"h{i}" for i in range(W.shape[0])]
    )
    transformations = pd.DataFrame(
        data={
            "psd": (np.power(weights, 2.0) / 2.0).mean(),
            "avg": weights.mean(),
            "med": weights.median(),
        }
    )
    transformations.plot(kind="bar", width=0.8, title=target_name, ax=axis)


def plot_features(
        results: Dict[str, MLPModel],
        n_rows, n_cols, all_columns
):
    sns.set_style("whitegrid")
    num_rows = n_rows
    num_cols = n_cols
    row, col = 0, 0
    _, ax = plt.subplots(num_rows, num_cols, figsize=(12, 14))
    for target in all_columns:
        plot_feature(results[target], axis=ax[row, col])
        col += 1
        if col == num_cols:
            row += 1
            col = 0

    plt.tight_layout()
    plt.show()


def layer_weights(dff_net, target, layer=0):
    W = extract_weights(dff_net.model)[layer]
    weights = pd.DataFrame(
        W, columns=dff_net.columns, index=[f"h{i}" for i in range(W.shape[0])]
    )
    transformations = pd.DataFrame(
        data={
            f"psd_{target}": (np.power(weights, 2.0) / 2.0).mean(),
            f"avg_{target}": weights.mean(),
            f"med_{target}": weights.median(),
        }
    )
    return transformations


def _plot_clusters(X: pd.DataFrame, K: DBSCAN, target: str, counts: Dict, min_counts=2):
    colors = [mpl.cm.tab20(i) for i in range(20)]
    fig, ax = plt.subplots(1, 3, figsize=(8, 2), sharey=True)
    minX, maxX = X.min().min(), X.max().max()
    minX -= (maxX - minX) * 0.1
    maxX += (maxX - minX) * 0.1
    for i in range(X.shape[0]):
        ax[0].scatter(X.iloc[i, 0], X.iloc[i, 1],
                      color=colors[K.labels_[i]], alpha=0.5)
        ax[1].scatter(X.iloc[i, 0], X.iloc[i, 2],
                      color=colors[K.labels_[i]], alpha=0.5)
        ax[2].scatter(X.iloc[i, 1], X.iloc[i, 2],
                      color=colors[K.labels_[i]], alpha=0.5)
        if counts[K.labels_[i]] <= min_counts:
            ax[0].text(
                X.iloc[i, 0],
                X.iloc[i, 1],
                f"{X.iloc[i,:].name}  ",
                horizontalalignment="right",
                verticalalignment="center",
            )
            ax[1].text(
                X.iloc[i, 0],
                X.iloc[i, 2],
                f"{X.iloc[i,:].name}  ",
                horizontalalignment="right",
                verticalalignment="center",
            )
            ax[2].text(
                X.iloc[i, 1],
                X.iloc[i, 2],
                f"{X.iloc[i,:].name}  ",
                horizontalalignment="right",
                verticalalignment="center",
            )
        ax[0].set_xlim((minX, maxX))
        ax[0].set_ylim((minX, maxX))
        ax[1].set_xlim((minX, maxX))
        ax[1].set_ylim((minX, maxX))
        ax[2].set_xlim((minX, maxX))
        ax[2].set_ylim((minX, maxX))
    fig.suptitle(target)


def summarize_weights(weights, feature_names, layer=0, scale=True):
    """
    Summarize the weights of a neural network model by calculating the mean,
    median, and positive semidefinite values of the weights for each feature.

    Parameters:
    - weights: The weights of the neural network model.
    - feature_names: A list of feature names.
    - layer: The layer of the neural network model from which to extract the weights.
    - scale: If True, scale the summary values.

    Returns:
    - psd: A DataFrame containing the summary values of the weights for each feature.

    """
    l1 = dict()
    psd = pd.DataFrame()
    for feature in feature_names:
        l1[feature] = layer_weights(weights[feature], feature, layer)
        psd = pd.concat(
            (psd, l1[feature]
             [[f"psd_{feature}", f"med_{feature}", f"avg_{feature}"]]),
            axis=1,
        )

    psd = psd.fillna(0)
    if scale:
        scaler = StandardScaler()
        psd = pd.DataFrame(scaler.fit_transform(psd),
                           index=psd.index, columns=psd.columns)
    return psd


def identify_relationships(weights, feature_names, eps=0.5, min_counts=2, plot=True):
    """
    Run a clustering algorithm on the summary values of weights coming out of input
    cells in the neural network. Summary values are the mean, the median and the
    positive semidefinite values. Those three dimensions are then clustered to
    identify what clusters have less or equal than min_count elements, to consider
    that cluster as relevant to produce the regression for that given feature the
    NN has been trained for.

    Parameters:
    - weights: The weights of the neural network model.
    - feature_names: A list of feature names.
    - eps: The maximum distance between two samples for one to be considered as in
        the neighborhood of the other.
    - min_counts: The minimum number of elements in a cluster to consider it relevant.
    - plot: If True, plot the clusters.

    Returns:
    - rels: A dictionary containing the relevant features for each target feature.

    """
    rels = {}
    for target in feature_names:
        X = weights[[f"psd_{target}", f"med_{target}",
                     f"avg_{target}"]].drop(target)
        K = DBSCAN(eps=eps, min_samples=1).fit(X)  # do_cluster(X, eps=eps)

        # Pairs with cluster_id: num_elements_per_cluster_id
        counts = Counter(K.labels_)

        rels[target] = []
        for i in range(X.shape[0]):
            if counts[K.labels_[i]] <= min_counts and X.iloc[i, :].name != "Noise":
                rels[target].append(X.iloc[i, :].name)
        if plot:
            _plot_clusters(X, K, target, counts)

    return rels


def _get_shap_values(model: MLPModel) -> np.ndarray:
    """
    Get the SHAP values for the given model.

    Parameters:
    - model: The model for which to compute the SHAP values.

    Returns:
    - shap_values: The SHAP values for the given model.

    """
    explainer = shap.DeepExplainer(
        model.model, model.train_loader.dataset.features)
    shap_values = explainer.shap_values(
        model.train_loader.dataset.features)

    return shap_values


def _average_shap_values(
        shap_values: Dict[str, np.ndarray],
        column_names: List[str],
        abs: bool = False) -> np.array:
    """
    Compute the average SHAP values for the given model.

    Parameters:
    - shap_values: The SHAP values for the given model.
    - column_names: The names of the columns in the dataset.
    - abs: If True, compute the absolute values of the SHAP values.

    Returns:
    - avg_shaps: The average SHAP values for the given model.

    """
    avg_shaps = []
    for i in range(len(column_names)):
        target_name = column_names[i]
        if abs:
            y = np.mean(np.abs(shap_values[target_name].T), axis=1) * 2.0
        else:
            y = np.mean(shap_values[target_name].T, axis=1) * 2.0
        avg_shaps.append(y)

    avg_shaps = np.array(avg_shaps)
    return avg_shaps


def _find_shap_elbow(avg_shaps: np.array, plot=False, verbose=False) -> float:
    n_bins = 20
    if avg_shaps.shape[0] < 20:
        n_bins = avg_shaps.shape[0]
    cutoff = 0
    # Try to find the elbow of the histogram counts unless the cutoff selected
    # is zero, in which case, the `knee` algorithm simply failed because
    # values in histogram are not monotonically decreasing uniformly.
    while cutoff == 0:
        histogram = np.histogram(avg_shaps, bins=n_bins)
        knee = kneed.KneeLocator(range(len(histogram[0])), histogram[0], S=1.0,
                                 curve="convex", direction="decreasing")
        threshold = histogram[1][knee.elbow]
        cutoff = knee.elbow
        if cutoff == 0:
            n_bins -= 1

    if plot:
        knee.plot_knee(figsize=(4, 3))
    if verbose:
        print(f"Cutoff pos.: {knee.elbow}; Threshold: {threshold:.4f}")

    return threshold


def _identify_edges(avg_shaps: np.array,
                    feature_names: List[str],
                    threshold: Dict[str, float]) -> Dict[str, List[Tuple[str, float]]]:
    rels = {}
    for i, target_name in enumerate(feature_names):
        labels = [f for f in feature_names if f != target_name]
        candidate_positions = np.argwhere(
            avg_shaps[i] > threshold[target_name]).flatten().tolist()
        rels[target_name] = [(labels[position], avg_shaps[i][position])
                             for position in candidate_positions]

    return rels


def _orient_edges_based_on_shap(G_shap: nx.DiGraph, verbose=False) -> nx.DiGraph:
    g = nx.DiGraph()
    already_checked = []
    if verbose:
        print("Orienting edges based on SHAP values ratio")
    for u, v, data in G_shap.edges(data=True):
        if G_shap.has_edge(v, u):
            if f"{u}->{v}" in already_checked:
                continue
            already_checked.append(f"{v}->{u}")
            reverse = G_shap.get_edge_data(v, u)
            diff = reverse['weight'] / data['weight']
            if verbose:
                print(f"{u}->{v}: {data['weight']:.3f} | ", end="")
                print(f"{v}->{u}: {reverse['weight']:.3f} | ")
                print(f"ratio: {diff:.2f}")
            if diff < 0.95:
                g.add_edge(u, v, weight=data['weight'])
            elif diff > 1.05:
                g.add_edge(v, u, weight=reverse['weight'])
            else:
                g.add_edge(u, v, weight=data['weight'])
                g.add_edge(v, u, weight=reverse['weight'])
        else:
            g.add_edge(u, v, weight=data['weight'])

    return g


def _remove_asymmetric_shap_edges(
    relationships: Dict[str, List[Tuple[str, float]]], verbose=False
) -> Dict[str, List[Tuple[str, float]]]:
    """
    Removes those edges who do not have a symmetric relationship found across the SHAP
    process.
    """
    shap_relationships = relationships.copy()
    asymmetric_edges = {}
    for key, values in shap_relationships.items():
        relation_names = [f"{relation[0]}" for relation in values]
        for potential_relation in relation_names:
            names = [e[0] for e in shap_relationships[potential_relation]]
            if key not in names:
                if key not in asymmetric_edges:
                    asymmetric_edges[key] = []
                asymmetric_edges[key].append(potential_relation)
                if verbose:
                    print(f"{key} -X-> {potential_relation} has NO symmetric")

    # Now remove those who are asymmetric.
    for target in shap_relationships.keys():
        if target in asymmetric_edges.keys():
            shap_relationships[target] = [
                tup
                for tup in shap_relationships[target]
                if tup[0] not in asymmetric_edges[target]
            ]

    return shap_relationships


def infer_causal_relationships(
        trained_models: Dict[str, MLPModel],
        feature_names: List[str],
        prune: bool = False,
        verbose=False,
        plot=False,
        prog_bar=True,
        silent=False):
    """
    Infer causal relationships between the input features and the target variable
    based on the SHAP values of the trained models.

    Parameters:
    - trained_models: A dictionary of trained models, where the keys are the target
        variable names and the values are the trained models.
    - feature_names: A list of input feature names.
    - prune: If True, remove asymmetric edges from the graph.
    - verbose: If True, print additional information.
    - plot: If True, plot the results.
    - prog_bar: If True, show a progress bar.
    - silent: If True, do not show any output.

    Returns:
    - A dictionary containing the SHAP values, the average SHAP values, the thresholds,
        the raw graph, and the oriented graph.

    """
    # pbar = tqdm(total=len(feature_names),
    #             **tqdm_params("Computing SHAPLEY values", prog_bar, silent=silent))
    if prog_bar and not silent:
        pbar = ProgBar().start_subtask(len(feature_names))

    shap_values = {}

    # desc="Computing SHAPLEY values",
    # disable=not prog_bar, position=1, leave=False)
    for target_name in feature_names:
        model = trained_models[target_name]
        shap_values[target_name] = _get_shap_values(model)
        pbar.update_subtask(1)

    avg_shaps = _average_shap_values(shap_values, feature_names)
    feature_threshold = {}
    for i, target in enumerate(feature_names):
        feature_threshold[target] = _find_shap_elbow(
            avg_shaps[i], plot, verbose)
    edges = _identify_edges(avg_shaps, feature_names, feature_threshold)
    if prune:
        edges = _remove_asymmetric_shap_edges(edges)
    G_shap = graph_from_dictionary(edges)
    G_shap_oriented = _orient_edges_based_on_shap(G_shap, verbose=False)

    return {
        'shap_values': shap_values,
        'avg_shaps': avg_shaps,
        'thresholds': feature_threshold,
        'raw_graph': G_shap,
        'graph': G_shap_oriented
    }



import numpy as np

from causalexplain.estimators.notears import main


def test_run_dispatches_to_variant(monkeypatch):
    def variant(data, loss, loss_grad, **kwargs):
        return {"called": True, "kwargs": kwargs}

    result = main.run(variant, data="data", loss="loss", loss_grad="grad", extra=1)
    assert result == {"called": True, "kwargs": {"extra": 1}}


def test_notears_standard_returns_solution(monkeypatch):
    monkeypatch.setattr(main.scipy.linalg, "expm",
                        lambda mat: np.eye(mat.shape[0]))

    def fake_minimize(func, w_flat, args=None, jac=None, method=None, options=None):
        class Res:
            def __init__(self, x):
                self.x = x
        return Res(w_flat)

    monkeypatch.setattr(main.scipy.optimize, "minimize", fake_minimize)

    data = np.array([[0.0, 0.0]])

    def simple_loss(W, data, cov, d, n):
        return float(np.sum(W ** 2))

    def simple_grad(W, data, cov, d, n):
        return 2 * W

    result = main.notears_standard(
        data=data,
        loss=simple_loss,
        loss_grad=simple_grad,
        max_iter=1,
        verbose=False,
    )
    assert "W" in result and "h" in result and "loss" in result
    assert result["W"].shape == (2, 2)
    assert result["h"] == 0.0


def test_notears_standard_can_return_progress(monkeypatch):
    monkeypatch.setattr(main.scipy.linalg, "expm",
                        lambda mat: np.eye(mat.shape[0]))
    monkeypatch.setattr(main.scipy.optimize, "minimize",
                        lambda func, w_flat, args=None, jac=None, method=None, options=None: type("Res", (), {"x": w_flat}))

    data = np.array([[0.0, 0.0]])
    progress = main.notears_standard(
        data=data,
        loss=lambda W, data, cov, d, n: 0.0,
        loss_grad=lambda W, data, cov, d, n: np.zeros_like(W),
        max_iter=1,
        output_all_progress=True)

    assert isinstance(progress, list)
    assert progress[0]["W"].shape == (2, 2)



import numpy as np
import pandas as pd
import pytest

from causalexplain.estimators.lingam import bootstrap


class DummyModel(bootstrap.BootstrapMixin):
    def __init__(self):
        self._causal_order = [0, 1]
        self._adjacency_matrix = np.array([[0.0, 0.0], [0.2, 0.0]])

    def fit(self, X):
        X = np.asarray(X)
        self._causal_order = list(range(X.shape[1]))
        self._adjacency_matrix = np.full((X.shape[1], X.shape[1]), 0.1)
        np.fill_diagonal(self._adjacency_matrix, 0.0)
        return self

    def estimate_total_effect(self, X, from_index, to_index):
        return float(from_index + to_index)


def test_bootstrap_validates_and_collects(monkeypatch):
    model = DummyModel()
    data = pd.DataFrame([[0.1, 0.2], [0.3, 0.4]])

    result = model.bootstrap(data, n_sampling=2)

    assert result.adjacency_matrices_.shape == (2, 2, 2)
    assert result.total_effects_.shape == (2, 2, 2)

    with pytest.raises(ValueError):
        model.bootstrap(data, n_sampling=0)
    with pytest.raises(ValueError):
        model.bootstrap(data.values, n_sampling="bad")


def test_causal_direction_counts_filters_by_threshold():
    ams = np.array([
        [[[0, 0.5], [0, 0]]],
        [[[0, -0.6], [0.1, 0]]],
    ]).reshape(2, 2, 2)
    res = bootstrap.BootstrapResult(ams, np.zeros_like(ams))

    counts = res.get_causal_direction_counts(min_causal_effect=0.4)
    assert counts["count"][0] == 2
    assert counts["from"][0] == 1 and counts["to"][0] == 0

    counts_split = res.get_causal_direction_counts(
        split_by_causal_effect_sign=True, min_causal_effect=0.4)
    assert "sign" in counts_split and len(counts_split["sign"]) == len(counts_split["from"])

    with pytest.raises(ValueError):
        res.get_causal_direction_counts(n_directions=0)


def test_directed_acyclic_graph_counts_and_probabilities():
    ams = np.array([
        [[0, 0.2], [0, 0]],
        [[0, -0.1], [0.4, 0]],
    ])
    res = bootstrap.BootstrapResult(ams, np.zeros_like(ams))

    dags = res.get_directed_acyclic_graph_counts(min_causal_effect=0.05)
    assert dags["count"][0] >= 1
    assert "from" in dags["dag"][0]

    dag_signs = res.get_directed_acyclic_graph_counts(
        min_causal_effect=0.05, split_by_causal_effect_sign=True)
    assert "sign" in dag_signs["dag"][0]

    probs = res.get_probabilities(min_causal_effect=0.1)
    assert probs.shape == (2, 2)

    ams_wide = np.array([[[0, 1, 0, 0], [0, 0, 0, 0]]])
    res_wide = bootstrap.BootstrapResult(ams_wide, np.zeros_like(ams_wide))
    splitted = res_wide.get_probabilities(min_causal_effect=0.0)
    assert len(splitted) == 2

    with pytest.raises(ValueError):
        res.get_directed_acyclic_graph_counts(n_dags=0)


def test_total_causal_effects_and_paths():
    total_effects = np.zeros((2, 2, 2))
    total_effects[0, 1, 0] = 0.5
    total_effects[1, 1, 0] = 0.8
    ams = np.array([
        [[0, 0.5], [0, 0]],
        [[0, 0.5], [0, 0]],
    ])
    res = bootstrap.BootstrapResult(ams, total_effects)

    te = res.get_total_causal_effects(min_causal_effect=0.0)
    assert te["from"] == [0]
    assert te["to"] == [1]
    assert te["effect"][0] == pytest.approx(0.65)
    assert te["probability"][0] == 1.0

    paths = res.get_paths(from_index=0, to_index=1)
    assert paths["path"][0] == [0, 1]
    assert paths["probability"][0] == 1.0

    with pytest.raises(ValueError):
        res.get_paths(0, 1, min_causal_effect=-1.0)



"""
Key changes and explanations:

1. Imported `numpy` for array operations and `train_lasso` from `train_lasso.py`.
2. Changed function signature to use Python conventions (e.g., `None` as default
for `pars`).
3. Used f-string for formatted output.
4. Converted `X` to a numpy array for easier indexing and shape retrieval.
5. Adjusted indexing to account for Python's 0-based indexing (e.g., `X[:, :k]`
instead of `X[,-k]`).
6. Implemented the selection vector creation using list comprehensions and boolean
operations.
7. Adjusted the `selVec` assignment to account for Python's slicing behavior.

Note that this translation assumes that the `train_lasso` function in `train_lasso.py`
returns a dictionary with a nested 'model' dictionary containing a 'beta' list.
You may need to adjust the `train_lasso` call and result handling if its
implementation differs from this assumption.
"""
import numpy as np

from .train_lasso import train_lasso


def selLasso(X, pars=None, output=False, k=None):
    if output:
        print(f"Performing variable selection for variable {k}:")

    result = {}
    X = np.asarray(X)
    p = X.shape

    if p[1] > 1:
        selVec = [False] * p[1]
        modfitGam = train_lasso(
            X[:, :k].tolist() + X[:, k+1:].tolist(), X[:, k].tolist(), pars)
        selVecTmp = [False] * (p[1] - 1)

        # Access the coefficients from the Lasso model
        coefficients = modfitGam['model'].coef_
        for i, coef in enumerate(coefficients):
            if coef != 0:
                selVecTmp[i] = True

        selVec[:k] = selVecTmp[:k]
        selVec[k+1:] = selVecTmp[k:]
    else:
        selVec = []

    return selVec



"""
Configuration for metric types used across the causalexplain package.
"""

# Non-causal metric types
global_nc_metric_types = [
    'r2', 'mse', 'rmse', 'mae', 'mape', 'medae',
    'evs', 'maxe', 'msle', 'male', 'mda', 'gmae'
]

# Causal metric types
global_metric_types = [
    'shd', 'precision', 'recall', 'f1', 'fdr', 'tpr',
    'fpr', 'nnz', 'extra', 'missing', 'reversed',
    'pred_size', 'pred_time', 'diff_edges'
]

__all__ = ['global_nc_metric_types', 'global_metric_types']



"""
StructureEstimator base class for PC and PC-stable algorithms.
"""
from functools import lru_cache
import pandas as pd


def convert_args_tuple(func):
    """
    Convert the arguments of a function to tuples.

    Args:
        obj: The object on which the function is called.
        variable: The variable of interest.
        parents: The parents of the variable.
        complete_samples_only: Flag indicating whether to use only complete samples.
        weighted: Flag indicating whether to use weighted samples.

    Returns:
        The result of the function with converted arguments.
    """
    def _convert_param_to_tuples(
        obj, variable, parents=tuple(), complete_samples_only=None, weighted=False
    ):
        parents = tuple(parents)
        return func(obj, variable, parents, complete_samples_only, weighted)

    return _convert_param_to_tuples


class StructureEstimator():
    """
    Base class for estimators in pgmpy; `ParameterEstimator`,
    `StructureEstimator` and `StructureScore` derive from this class.
    """

    data = None
    complete_samples_only = True
    variables = None
    state_names = None

    def __init__(self, independencies=None):
        self.independencies = independencies
        if self.independencies is not None:
            self.variables = self.independencies.get_all_variables()

    def _init_data(self, data=None, state_names=None, complete_samples_only=True):
        """
        Initialize the data for the estimator.

        Parameters
        ----------
        data: pandas DataFrame object datafame object where each column represents one variable.
            (If some values in the data are missing the data cells should be set to `numpy.NaN`.
            Note that pandas converts each column containing `numpy.NaN`s to dtype `float`.)

        state_names: dict (optional)
            A dict indicating, for each variable, the discrete set of states (or values)
            that the variable can take. If unspecified, the observed values in the data set
            are taken to be the only possible states.

        complete_samples_only: bool (optional, default `True`)
            Specifies how to deal with missing data, if present. If set to `True` all rows
            that contain `np.Nan` somewhere are ignored. If `False` then, for each variable,
            every row where neither the variable nor its parents are `np.NaN` is used.
            This sets the behavior of the `state_count`-method.
        """
        self.data = data
        # data can be None in the case when learning structre from
        # independence conditions. Look into PC.py.
        if self.data is not None:
            self.complete_samples_only = complete_samples_only

            self.variables = list(data.columns.values)

            if not isinstance(state_names, dict):
                self.state_names = {
                    var: self._collect_state_names(var) for var in self.variables
                }
            else:
                self.state_names = dict()
                for var in self.variables:
                    if var in state_names:
                        if not set(self._collect_state_names(var)) <= set(
                            state_names[var]
                        ):
                            raise ValueError(
                                f"Data contains unexpected states for variable: {var}."
                            )
                        self.state_names[var] = state_names[var]
                    else:
                        self.state_names[var] = self._collect_state_names(var)

    def _collect_state_names(self, variable):
        "Return a list of states that the variable takes in the data."
        states = sorted(list(self.data.loc[:, variable].dropna().unique()))
        return states

    @convert_args_tuple
    @lru_cache(maxsize=2048)
    def state_counts(
        self, variable, parents=None, complete_samples_only=None, weighted=False
    ):
        """
        Return counts how often each state of 'variable' occurred in the data.
        If a list of parents is provided, counting is done conditionally
        for each state configuration of the parents.

        Parameters
        ----------
        variable: string
            Name of the variable for which the state count is to be done.

        parents: list
            Optional list of variable parents, if conditional counting is desired.
            Order of parents in list is reflected in the returned DataFrame

        complete_samples_only: bool
            Specifies how to deal with missing data, if present. If set to `True` all rows
            that contain `np.NaN` somewhere are ignored. If `False` then
            every row where neither the variable nor its parents are `np.NaN` is used.
            Desired default behavior can be passed to the class constructor.

        weighted: bool
            If True, data must have a `_weight` column specifying the weight of the
            datapoint (row). If False, each datapoint has a weight of `1`.

        Returns
        -------
        state_counts: pandas.DataFrame
            Table with state counts for 'variable'

        Examples
        --------
        >>> import pandas as pd
        >>> from pgmpy.estimators import BaseEstimator
        >>> data = pd.DataFrame(data={'A': ['a1', 'a1', 'a2'],
                                      'B': ['b1', 'b2', 'b1'],
                                      'C': ['c1', 'c1', 'c2']})
        >>> estimator = BaseEstimator(data)
        >>> estimator.state_counts('A')
            A
        a1  2
        a2  1
        >>> estimator.state_counts('C', parents=['A', 'B'])
        A  a1      a2
        B  b1  b2  b1  b2
        C
        c1  1   1   0   0
        c2  0   0   1   0
        >>> estimator.state_counts('C', parents=['A'])
        A    a1   a2
        C
        c1  2.0  0.0
        c2  0.0  1.0
        """
        if parents is None:
            parents = []
        parents = list(parents)

        # default for how to deal with missing data can be set in class constructor
        if complete_samples_only is None:
            complete_samples_only = self.complete_samples_only
        # ignores either any row containing NaN, or only those where the variable
        # or its parents is NaN
        data = (
            self.data.dropna()
            if complete_samples_only
            else self.data.dropna(subset=[variable] + parents)
        )

        if weighted and ("_weight" not in self.data.columns):
            raise ValueError(
                "data must contain a `_weight` column if weighted=True")

        if not parents:
            # count how often each state of 'variable' occured
            if weighted:
                state_count_data = data.groupby([variable]).sum()["_weight"]
            else:
                state_count_data = data.loc[:, variable].value_counts()

            state_counts = (
                state_count_data.reindex(self.state_names[variable])
                .fillna(0)
                .to_frame()
            )
            # Ensure the column label is the variable name, regardless of the series'
            # original name (pandas version differences can leave it as None or 0).
            state_counts.columns = [variable]

        else:
            parents_states = [self.state_names[parent] for parent in parents]
            # count how often each state of 'variable' occured, conditional on
            # parents' states
            if weighted:
                state_count_data = (
                    data.groupby(
                        [variable] + parents).sum()["_weight"].unstack(parents)
                )

            else:
                state_count_data = (
                    data.groupby([variable] + parents).size().unstack(parents)
                )

            if not isinstance(state_count_data.columns, pd.MultiIndex):
                state_count_data.columns = pd.MultiIndex.from_arrays(
                    [state_count_data.columns]
                )

            # reindex rows & columns to sort them and to add missing ones
            # missing row    = some state of 'variable' did not occur in data
            # missing column = some state configuration of current 'variable's parents
            #                  did not occur in data
            row_index = self.state_names[variable]
            column_index = pd.MultiIndex.from_product(
                parents_states, names=parents)
            state_counts = state_count_data.reindex(
                index=row_index, columns=column_index
            ).fillna(0)

        return state_counts



"""
This Python translation aims to maintain the functionality of the original R code. 
Here are some key points about the translation:

- The function signature remains similar, with default values for `pars` and `output`.
- We use numpy arrays instead of R matrices.
- The `xselect()` method is assumed to be part of the model returned by 
  `train_LMboost`. You may need to adjust this based on the actual implementation.
- List comprehensions and numpy operations are used to replace some R-specific 
  operations.
- The indexing is adjusted to account for Python's 0-based indexing (compared to R's 
  1-based indexing).

Note that this translation assumes that the `train_LMboost` function in Python returns 
a dictionary with a 'model' key, which has an `xselect()` method. You may need to 
adjust this part based on the actual implementation of `train_LMboost` in Python.
"""
import numpy as np

from Python.CAM.train_LMboost import train_LMboost


def selLmBoost(X, pars=None, output=False, k: int = None):
    if pars is None:
        pars = {'atLeastThatMuchSelected': 0.02, 'atMostThatManyNeighbors': 10}

    if output:
        print(f"Performing variable selection for variable {k}:")

    X = np.array(X)
    p = X.shape

    if p[1] > 1:
        selVec = [False] * p[1]

        modfit_lm = train_LMboost(np.delete(X, k, axis=1), X[:, k], pars)
        # Get the indices of the important features
        cc = np.argsort(modfit_lm['model'].feature_importances_)[::-1]

        if output:
            print("The following variables")
            print(cc)

        # Create a list to store the importance of each feature in 'cc'
        how_often_selected = modfit_lm['model'].feature_importances_

        if output:
            print("... have been selected that many times:")
            print(how_often_selected)

        how_often_selected_sorted = sorted(how_often_selected, reverse=True)

        if sum(np.array(how_often_selected) > pars['atLeastThatMuchSelected']) \
                > pars['atMostThatManyNeighbors']:
            cc = cc[np.array(how_often_selected) >
                    how_often_selected_sorted[pars['atMostThatManyNeighbors']]]
        else:
            cc = cc[np.array(how_often_selected) >
                    pars['atLeastThatMuchSelected']]

        if output:
            print("We finally choose as possible parents:")
            print(cc)
            print()

        tmp = [False] * (p[1] - 1)
        for idx in cc:
            tmp[idx] = True

        selVec[:k] = tmp[:k]
        selVec[k+1:] = tmp[k:]
    else:
        selVec = []

    return selVec



import networkx as nx
import pytest

from causalexplain.independence.cond_indep import (
    ConditionalIndependencies,
    SufficientSets,
    get_backdoor_paths,
    get_conditional_independencies,
    get_paths,
    get_sufficient_sets,
    get_sufficient_sets_for_pair,
    find_colliders_in_path,
)


def test_conditional_independencies_formats_and_deduplicates():
    ci = ConditionalIndependencies()
    assert repr(ci) == "{}"

    ci.add("A", "B")
    ci.add("B", "C", ["D"])
    ci.add("B", "C", ["D"])  # duplicate

    text = str(ci)
    assert "A ⊥ B" in text
    assert "B ⊥ C | ('D',)" in text
    with pytest.raises(ValueError):
        repr(ci)


def test_sufficient_sets_add_and_repr_and_str():
    ss = SufficientSets()
    assert str(ss) == "No sufficient sets found"
    assert repr(ss) == "[]"

    ss.add([("X", "Y"), ("A", "B")])
    ss.add([("X", "Y")])  # already present
    rendered = str(ss)
    assert "('X', 'Y')" in rendered and "('A', 'B')" in rendered
    assert repr(ss).startswith("[('X', 'Y'), ('A', 'B')]")


@pytest.mark.parametrize(
    "x,y,expected",
    [
        ("x", "y", []),  # nodes missing
        ("x", "x", []),  # identical nodes
    ],
)
def test_get_backdoor_paths_handles_missing_and_same_nodes(x, y, expected):
    dag = nx.DiGraph()
    dag.add_edges_from([("a", "b"), ("b", "c")])
    assert get_backdoor_paths(dag, x, y) == expected


def test_get_backdoor_paths_returns_only_paths_starting_towards_source():
    dag = nx.DiGraph()
    dag.add_edges_from([("m", "x"), ("m", "z"), ("z", "y"), ("k", "x")])
    # Only paths where the first edge aims into x qualify
    paths = get_backdoor_paths(dag, "x", "y")
    assert paths == [["x", "m", "z", "y"]]


@pytest.mark.parametrize(
    "x,y,expected",
    [
        ("x", "y", []),
        ("x", "x", []),
        ("x", "z", [["x", "z"]]),
    ],
)
def test_get_paths_branches(x, y, expected):
    dag = nx.DiGraph()
    dag.add_edges_from([("x", "z")])
    assert get_paths(dag, x, y) == expected


def test_find_colliders_in_path_varied_cases():
    dag = nx.DiGraph()
    dag.add_edges_from([("a", "b"), ("c", "b"), ("b", "d")])
    assert find_colliders_in_path(dag, ["a", "b", "d"]) == set()
    assert find_colliders_in_path(dag, ["a", "b", "c"]) == {"b"}
    assert find_colliders_in_path(dag, ["a", "b"]) == set()


def test_get_sufficient_sets_for_pair_filters_descendants_and_colliders():
    dag = nx.DiGraph()
    dag.add_edges_from([("z", "x"), ("z", "y"), ("x", "z2"), ("z2", "y")])
    assert get_sufficient_sets_for_pair(dag, "x", "y") == [["z"]]

    # Introduce collider so sets are discarded
    dag.add_edge("y", "z")
    assert get_sufficient_sets_for_pair(dag, "x", "y") == []


def test_get_sufficient_sets_collects_all_pairs():
    dag = nx.DiGraph()
    dag.add_edges_from([("z", "x"), ("z", "y")])
    suff_sets = get_sufficient_sets(dag)
    assert ["z"] in suff_sets._cache


def test_get_conditional_independencies_tracks_paths_and_colliders():
    dag = nx.DiGraph()
    dag.add_edges_from([("x", "z"), ("z", "y")])
    cond_indeps = get_conditional_independencies(dag, verbose=False)

    # x and y are connected through z with no colliders -> conditioning on z
    assert ("x", "y", ("z",)) in cond_indeps._cache

    # Nodes without open paths become unconditioned independencies
    isolated = nx.DiGraph()
    isolated.add_nodes_from(["a", "b"])
    empty_cond = get_conditional_independencies(isolated, verbose=False)
    assert ("a", "b", None) in empty_cond._cache



"""
Permutation Importance for feature selection.
Wrapper over SKLearn's PermutationImportance and own implementation of
the vanilla version of the algorithm to run over models trained with PyTorch.

(C) J. Renero 2022, 2023

Parameters:
-----------
models: dict
    A dictionary of models, where the keys are the target variables and
    the values are the models trained to predict the target variables.
n_repeats: int
    The number of times to repeat the permutation importance algorithm.
mean_pi_percentile: float
    The percentile of the mean permutation importance to use as a threshold
    for feature selection.
random_state: int
    The random state to use for the permutation importance algorithm.
prog_bar: bool
    Whether to display a progress bar or not.
verbose: bool
    Whether to display explanations on the process or not.
silent: bool
    Whether to display anything or not.

"""

# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches

from typing import List, Tuple

import networkx as nx
import numpy as np
import pandas as pd
import torch
from matplotlib import pyplot as plt
from mlforge.progbar import ProgBar
from sklearn.base import BaseEstimator
from sklearn.inspection import permutation_importance
from sklearn.preprocessing import StandardScaler

from causalexplain.common import utils
from causalexplain.common.plot import subplots
from causalexplain.explainability.hierarchies import Hierarchies
from causalexplain.independence.feature_selection import select_features
from causalexplain.models._models import MLPModel


class PermutationImportance(BaseEstimator):
    """
    Permutation Importance for feature selection.
    Wrapper over SKLearn's PermutationImportance and own implementation of
    the vanilla version of the algorithm to run over models trained with PyTorch.
    """

    device = utils.select_device("cpu")

    def __init__(
            self,
            models: dict,
            discrepancies: dict = None,
            correlation_th: float = None,
            n_repeats: int = 10,
            mean_pi_percentile: float = 0.8,
            exhaustive: bool = False,
            threshold: float = None,
            random_state: int = 42,
            prog_bar=True,
            verbose=False,
            silent=False):
        """
        Parameters:
        -----------
        models: dict
            A dictionary of models, where the keys are the target variables and
            the values are the models trained to predict the target variables.
        discrepancies: dict
            A dictionary of discrepancies for each target variable, based on SHAP
            values.
        n_repeats: int
            The number of times to repeat the permutation importance algorithm.
        mean_pi_percentile: float
            The percentile of the mean permutation importance to use as a threshold
            for feature selection.
        random_state: int
            The random state to use for the permutation importance algorithm.
        prog_bar: bool
            Whether to display a progress bar or not.
        verbose: bool
            Whether to display explanations on the process or not.
        silent: bool
            Whether to display anything or not.
        """
        super().__init__()
        self.models = models
        self.shap_discrepancies = discrepancies
        self.regressors = models.regressor
        self.correlation_th = correlation_th
        self.n_repeats = n_repeats
        self.mean_pi_percentile = mean_pi_percentile
        self.exhaustive = exhaustive
        self.threshold = threshold
        self.random_state = random_state
        self.prog_bar = prog_bar
        self.verbose = verbose
        self.silent = silent
        self.feature_names = list(self.regressors.keys())

        self.base_loss = {}
        self.base_std = {}
        self.all_pi = []
        self.pi = {}
        self.connections = {}
        self.corr_matrix = None
        self.correlated_features = None
        self.G_pi = None
        self.mean_pi_threshold = None

        self.is_fitted_ = False

        self._fit_desc = "Running Perm.Importance"
        self._pred_desc = "Predicting w perm. imp."

    def fit(self, X):
        """
        Implementation of the fit method for the PermutationImportance class.
        If the model is a PyTorch model, the fit method will compute the base loss
        for each feature. If the model is a SKLearn model, the fit method will
        compute the permutation importance for each feature.
        """
        self._obtain_correlation_info(X)

        first_key = self.feature_names[0]
        if isinstance(self.regressors[first_key], MLPModel):
            return self._fit_pytorch()
        else:
            return self._fit_sklearn(X)

    def _obtain_correlation_info(self, X):
        if self.correlation_th:
            self.corr_matrix = Hierarchies.compute_correlation_matrix(X)
            self.correlated_features = Hierarchies.compute_correlated_features(
                self.corr_matrix, self.correlation_th, self.feature_names,
                verbose=self.verbose)

    def _fit_pytorch(self):
        """
        Fit the model to compute the base loss for each feature for pyTorch models.
        """
        pbar = ProgBar().start_subtask("Perm.Imp_fit", len(self.feature_names))
        print("Computing base loss (PyTorch)") if self.verbose else None

        for feature_idx, feature in enumerate(self.feature_names):
            print(f"Feature: {feature} ", end="") if self.verbose else None

            regressor = self.regressors[feature]
            model = regressor.model.to(self.device)

            avg_loss, std_loss, _ = self._compute_loss_shuffling_column(
                model, regressor.train_loader)
            self.base_loss[feature] = avg_loss
            self.base_std[feature] = std_loss

            if (self.verbose) and (not self.silent):
                print(f"Base loss: {self.base_loss[feature]:.6f} ", end="")
                print(f"+/- {self.base_std[feature]:.6f}")

            pbar.update_subtask("Perm.Imp_fit", feature_idx + 1)

        pbar.remove("Perm.Imp_fit")
        self.is_fitted_ = True

        return self

    def _fit_sklearn(self, X):
        """
        Fit the model to compute the base loss for each feature, for SKLearn models.
        """
        pbar = ProgBar().start_subtask("Perm.Imp_fit(sklearn)", len(self.feature_names))

        # If me must exclude features due to correlation, we must do it before
        # computing the base loss
        if self.correlation_th:
            self.corr_matrix = Hierarchies.compute_correlation_matrix(X)
            self.correlated_features = Hierarchies.compute_correlated_features(
                self.corr_matrix, self.correlation_th, self.feature_names,
                verbose=self.verbose)

        self.pi = {}
        self.all_pi = []
        X_original = X.copy()
        for target_idx, target_name in enumerate(self.feature_names):
            X = X_original.copy()

            # if correlation_th is not None then, remove features that are highly
            # correlated with the target, at each step of the loop
            if self.correlation_th is not None:
                if len(self.correlated_features[target_name]) > 0:
                    X = X.drop(self.correlated_features[target_name], axis=1)
                    if self.verbose:
                        print("REMOVED CORRELATED FEATURES: ",
                              self.correlated_features[target_name])

            # print(f"Feature: {target_name} ", end="") if self.verbose else None

            regressor = self.regressors[target_name]
            y = X[target_name]
            X = X.drop(columns=[target_name])
            self.pi[target_name] = permutation_importance(
                regressor, X, y, n_repeats=10,
                random_state=self.random_state)

            if self.correlation_th is not None:
                self._add_zeroes(
                    target_name, self.correlated_features[target_name])

            self.all_pi.append(self.pi[target_name]['importances_mean'])

            pbar.update_subtask("Perm.Imp_fit(sklearn)", target_idx + 1)

        pbar.remove("Perm.Imp_fit(sklearn)")

        self.all_pi = np.array(self.all_pi).flatten()
        self.mean_pi_threshold = np.quantile(
            self.all_pi, self.mean_pi_percentile)
        self.is_fitted_ = True

        return self

    def predict(self, X=None, root_causes=None, prior: List[List[str]] = None):
        """
        Implementation of the predict method for the PermutationImportance class.

        Parameters:
        -----------
        X: pd.DataFrame
            The data to predict the permutation importance for.

        Returns:
        --------
        G_pi: nx.DiGraph
            The DAG representing the permutation importance for the features.
        """
        if self.verbose:
            print("-----\npermutation_importance.predict()")

        self.prior = prior
        first_key = self.feature_names[0]
        if isinstance(self.regressors[first_key], MLPModel):
            return self._predict_pytorch(X, root_causes)

        #  SKLearn models don't have a predict stage for permutation importance.
        return self._predict_sklearn(X, root_causes)

    def _predict_pytorch(self, X, root_causes) -> nx.DiGraph:
        """
        Predict the permutation importance for each feature, for each target,
        under the PyTorch implementation of the algorithm.
        """
        pbar = ProgBar().start_subtask("Perm.Imp_predict", len(self.feature_names))
        print("Computing permutation loss (PyTorch)") if self.verbose else None

        self.all_pi = []
        num_vars = len(self.feature_names)
        for target_idx, target in enumerate(self.feature_names):
            regressor = self.regressors[target]
            model = regressor.model
            feature_names_wo_target = [
                f for f in self.feature_names if f != target]
            candidate_causes = utils.valid_candidates_from_prior(
                self.feature_names, target, self.prior)
            if self.verbose:
                print(
                    f"Target: {target} (base loss: {self.base_loss[target]:.6f})")

            # Create the dictionary to store the permutation importance, same way
            # as the sklearn implementation
            self.pi[target] = {}
            if self.correlation_th is not None:
                num_vars = len(self.feature_names) - \
                    len(self.correlated_features[target])
                # Filter out features that are highly correlated with the target
                candidate_causes = [f for f in candidate_causes
                                    if f not in self.correlated_features[target]]

            # Compute the permutation importance for each feature
            self.pi[target]['importances_mean'], self.pi[target]['importances_std'] = \
                self._compute_perm_imp(target, regressor, model, num_vars)


            if self.correlation_th is not None:
                self._add_zeroes(target, self.correlated_features[target])

            self.all_pi.append(self.pi[target]['importances_mean'])

            self.connections[target] = select_features(
                values=self.pi[target]['importances_mean'],
                feature_names=feature_names_wo_target,
                exhaustive=self.exhaustive,
                threshold=self.mean_pi_threshold,
                verbose=self.verbose)

            pbar.update_subtask("Perm.Imp_predict", target_idx)

        pbar.remove("Perm.Imp_predict")
        self.G_pi = self._build_pi_dag(X, root_causes)
        self.GP_pi = utils.break_cycles_if_present(
            self.G_pi, self.shap_discrepancies, self.prior, verbose=self.verbose)
        return self.G_pi

    def _predict_sklearn(self, X, root_causes) -> nx.DiGraph:
        """
        Predict the permutation importance for each feature, for each target,
        under the PyTorch implementation of the algorithm.
        """
        print("Computing permutation loss (SKLearn)") if self.verbose else None

        self.connections = {}
        for target in self.feature_names:
            if self.verbose:
                print(f"Target: {target} ")
            candidate_causes = [f for f in self.feature_names if f != target]
            self.connections[target] = select_features(
                values=self.pi[target]['importances_mean'],
                feature_names=candidate_causes,
                verbose=self.verbose)

        self.G_pi = self._build_pi_dag(X, root_causes)
        return self.G_pi

    def _build_pi_dag(self, X, root_causes):
        """
        Build a DAG from the permutation importance results. This is the last stage
        of the algorithm, where we build a DAG from the permutation importance results.
        Placed in a separate method as it is shared in SKLearn and PyTorch.
        """
        self.G_pi = utils.digraph_from_connected_features(
            X, self.feature_names, self.models, self.connections, root_causes,
            reciprocity=True, anm_iterations=10, verbose=self.verbose)

        self.all_pi = np.array(self.all_pi).flatten()
        self.mean_pi_threshold = np.quantile(
            self.all_pi, self.mean_pi_percentile)

        return self.G_pi

    def _compute_perm_imp(self, target, regressor, model, num_vars):
        """
        Compute the permutation importance for each feature, for a given target
        variable.

        Parameters:
        -----------
        target: str
            The target variable to compute the permutation importance for.
        regressor: MLPModel
            The regressor to compute the permutation importance for.
        model: torch.nn.Module
            The model to compute the permutation importance for.
        num_vars: int
            The number of variables to compute the permutation importance for.

        Returns:
        --------
        importances_mean: np.ndarray
            The mean permutation importance for each feature.
        importances_std: np.ndarray
            The standard deviation of the permutation importance for each feature.
        """
        importances_mean = []
        importances_std = []
        for shuffle_col in range(num_vars-1):
            feature = regressor.columns[shuffle_col]
            print(
                f"  ↳ Feature: {feature} ", end="") if self.verbose else None

            _, _, losses = self._compute_loss_shuffling_column(
                model, regressor.train_loader, shuffle_col=shuffle_col)

            axis = 1 if self.n_repeats > 1 else 0
            perm_importances = np.mean(
                losses, axis=axis) - self.base_loss[target]

            importances_mean.append(
                np.mean(perm_importances))
            if self.n_repeats > 1:
                importances_std.append(np.std(perm_importances))
            else:
                importances_std.append(
                    np.abs(np.std(losses) - self.base_loss[target]))

            if self.verbose:
                print(
                    f"Perm.imp.: {importances_mean[-1]:.6f} "
                    f"+/- {importances_std[-1]:.6f}")

        return np.array(importances_mean), np.array(importances_std)

    def _add_zeroes(self, target, correlated_features):
        """
        Add zeroes to the mean perm imp. values for correlated features.
        """
        features = [f for f in self.feature_names if f != target]
        for correlated_feature in correlated_features:
            correlated_feature_position = features.index(correlated_feature)
            self.pi[target]['importances_mean'] = np.insert(
                self.pi[target]['importances_mean'], correlated_feature_position, 0.)
            self.pi[target]['importances_std'] = np.insert(
                self.pi[target]['importances_std'], correlated_feature_position, 0.)

    def fit_predict(self, X, root_causes):
        self._obtain_correlation_info(X)

        first_key = self.feature_names[0]
        if isinstance(self.regressors[first_key], MLPModel):
            return self._fit_predict_pytorch(X, root_causes)
        else:
            return self._fit_predict_sklearn(X)

    def _fit_predict_pytorch(self, X, root_causes):
        self._fit_pytorch()
        return self._predict_pytorch(X, root_causes)

    def _fit_predict_sklearn(self, X):
        self._fit_sklearn(X)
        return self.pi

    def _compute_loss_shuffling_column(
            self,
            model: torch.nn.Module,
            dataloader: torch.utils.data.DataLoader,
            shuffle_col: int = -1) -> Tuple[float, float, np.ndarray]:
        """
        Computes the average MSE loss for a given model and dataloader.

        Parameters:
        -----------
        model: torch.nn.Module
            The model to compute the loss for.
        dataloader: torch.utils.data.DataLoader
            The dataloader to use for computing the loss.
        shuffle: int
            If > 0, the column of the input data to shuffle.

        Returns:
        --------
        avg_loss: float
            The average MSE loss.
        std_loss: float
            The standard deviation of the MSE loss.
        losses: np.ndarray
            The MSE loss for each batch.
        """
        mse = np.array([])
        num_batches = 0
        print(
            f"(Repeats: {self.n_repeats}) ", end="") if self.verbose else None
        for _ in range(self.n_repeats):
            loss = []
            # Loop over all batches in train loader
            for _, (X, y) in enumerate(dataloader):
                X = X.to(self.device)
                y = y.to(self.device)

                # Shuffle data if specified
                if shuffle_col >= 0:
                    X = self._shuffle_2Dtensor_column(X, shuffle_col)

                # compute MSE loss for each batch
                yhat = model.forward(X)
                loss.append(model.loss_fn(yhat, y).item())
                num_batches += 1

            if len(mse) == 0:
                mse = np.array(loss)
            else:
                mse = np.vstack((mse, [loss]))

        return np.mean(mse), np.std(mse), mse

    def _shuffle_2Dtensor_column(
            self,
            tensor: torch.Tensor,
            column: int) -> torch.Tensor:
        """
        Shuffles the column of a 2D tensor.

        Parameters:
        -----------
        tensor: torch.Tensor
            The tensor to shuffle.
        column: int
            The column to shuffle.

        Returns:
        --------
        shuffled_tensor: torch.Tensor
            The shuffled tensor.
        """
        assert column < tensor.shape[1], "Column index out of bounds"
        assert len(tensor.shape) == 2, "Tensor must be 2D"
        num_rows, num_columns = tensor.shape
        idx = torch.randperm(tensor.shape[0], device=self.device)
        column_reshuffled = torch.reshape(tensor[idx, column], (num_rows, 1))
        if column == 0:
            return torch.cat((column_reshuffled, tensor[:, 1:]), 1)
        else:
            return torch.cat((tensor[:, 0:column], column_reshuffled, tensor[:, column+1:]), 1)

    def plot(self, **kwargs):
        """
        Plot the permutation importance for each feature, by calling the internal
        _plot_perm_imp method.

        Parameters:
        -----------
        kwargs: dict
            Keyword arguments to pass to the _plot_perm_imp method.
            Examples:
                - figsize: tuple

        Returns:
        --------
        fig: matplotlib.figure.Figure
            The figure containing the plot.
        """
        assert self.is_fitted_, "Model not fitted yet"
        plot_args = [(target_name) for target_name in self.feature_names]
        return subplots(self._plot_perm_imp, *plot_args, **kwargs)

    def _plot_perm_imp(self, target, ax, **kwargs):
        """
        Plot the permutation importance for a given target variable.

        Parameters:
        -----------
        target: str
            The target variable to plot the permutation importance for.
        ax: matplotlib.axes.Axes
            The axes to plot the permutation importance on.
        kwargs: dict
            Keyword arguments to pass to the barh method of the axes.
            Examples:
                - figsize: tuple

        Returns:
        --------
        fig: matplotlib.figure.Figure
            The figure containing the subplot.

        """
        feature_names = [f for f in self.feature_names if f != target]
        figsize_ = kwargs.get('figsize', (6, 3))
        fig = None
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize_)

        sorted_idx = self.pi[target]['importances_mean'].argsort()
        ax.barh(
            np.array(feature_names)[sorted_idx.astype(int)],
            self.pi[target]['importances_mean'][sorted_idx],
            xerr=self.pi[target]['importances_std'][sorted_idx],
            align='center', alpha=0.5)

        xlims = ax.get_xlim()
        if xlims[1] < self.mean_pi_threshold:
            ax.set_xlim(right=self.mean_pi_threshold +
                        ((xlims[1] - xlims[0])/20))
        ax.axvline(
            x=self.mean_pi_threshold, color='red', linestyle='--', linewidth=0.5)

        ax.set_title(
            f"Perm.Imp.{target}: {','.join(self.connections[target])}")
        fig = ax.figure if fig is None else fig

        return fig


if __name__ == "__main__":
    path = "/Users/renero/phd/data/RC3/"
    output_path = "/Users/renero/phd/output/RC3/"
    experiment_name = 'rex_generated_gp_mix_1'

    ref_graph = utils.graph_from_dot_file(f"{path}{experiment_name}.dot")
    data = pd.read_csv(f"{path}{experiment_name}.csv")
    scaler = StandardScaler()
    data = pd.DataFrame(scaler.fit_transform(data), columns=data.columns)
    rex = utils.load_experiment(f"{experiment_name}_gbt", output_path)
    print(f"Loaded experiment {experiment_name}")

    #  Run the permutation importance algorithm
    pi = PermutationImportance(
        rex.models, n_repeats=10, prog_bar=False, verbose=True)
    pi.fit(data)
    pi.predict(data, rex.root_causes)
    pi.plot(figsize=(7, 5))
    plt.show()
